from . import base_ctx
from .modbus_protocol_client import build_req_read_words, parse_res_tcp


class MbCtx(base_ctx.BaseCtx):
    def __init__(self, channel):
        super().__init__(channel)
        self.log = None
        self.trans_id_ = 0
        self.request_ = None

    def data_cb(self, data):
        return parse_res_tcp(data, 0, len(data), self.request_)

    # returns:
    #  - bytearray() with words
    #  * MbException
    #  * MbError
    #  * ConnectionError, TimeoutError
    def read_words(self, slave_nb: int, fn_nb: int, addr: int, nb_words: int, comment: str, log):
        self.trans_id_ = (self.trans_id_ + 1) & 0xffff
        self.request_ = (self.trans_id_, slave_nb, fn_nb, addr, nb_words)

        req = build_req_read_words(self.trans_id_, slave_nb, fn_nb, addr, nb_words)

        trans_id, slave_nb, fn_nb, data_bytes = self.send_rcv_raw(req, comment, log)        # pylint: disable=unused-variable

        return data_bytes



#!/usr/bin/python3

MIN_I16 = -0x8000
MAX_I16 = 0x7fff
MAX_UI16 = 0xffff

MIN_I32 = -0x80000000
MAX_I32 = 0x7fffffff
MAX_UI32 = 0xffffffff

MIN_I64 = -0x8000000000000000
MAX_I64 = 0x7fffffffffffffff
MAX_UI64 = 0xffffffffffffffff




class INamedValues:
    def contains(self, name: str) -> bool:
        raise NotImplementedError()

    # None if does not contain
    def get_raw_value(self, name: str):
        raise NotImplementedError()


class ILocation:
    def get_location(self) -> (str, None):
        raise NotImplementedError()


class StrLocation(ILocation):
    def __init__(self, s: str):
        assert isinstance(s, str), s
        self.str_ = s

    def get_location(self) -> (str, None):
        return self.str_


class NamedParameters:
    def __init__(self, src: INamedValues, exc_type: type = Warning, location: ILocation = None):
        assert isinstance(src, INamedValues), src
        assert isinstance(exc_type, type), exc_type
        assert location is None or isinstance(location, ILocation), location

        self.src_ = src
        self.exc_type_ = exc_type
        self.location_ = location

    def _exc(self, msg: str) -> Exception:
        if self.location_:
            loc = self.location_.get_location()
            if loc:
                assert isinstance(loc, str), loc
                return self.exc_type_(loc + ': ' + msg)

        return self.exc_type_(msg)

    def get_raw(self, name: str, allow_none: bool = False, optional: bool = False, default=None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional

        if not self.src_.contains(name):
            if optional:
                return default
            raise self._exc('Parameter {} is missing'.format(name))

        v = self.src_.get_raw_value(name)
        if v is None and not allow_none:
            raise self._exc('Parameter {} is None'.format(name))
        return v

    # options should consist of str. None should not be put there
    def get_str(self, name: str, allow_none: bool = False, optional: bool = False, default: str = None, options: (list, tuple) = None) -> (str, None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional
        assert default is None or isinstance(default, str), default
        assert options is None or isinstance(options, (list, tuple)), options

        if not self.src_.contains(name):
            if optional:
                return default
            raise self._exc('Parameter {} is missing'.format(name))

        v = self.src_.get_raw_value(name)
        if v is None:
            if allow_none:
                return None
            raise self._exc('Parameter {} is None'.format(name))

        if not isinstance(v, str):
            raise self._exc('Parameter {}={} should be str, not {}'.format(name, v, type(v)))

        if options and v not in options:
            raise self._exc('Parameter {}={} is invalid, available values: {}'.format(name, v, options))

        return v

    def get_bool(self, name: str, allow_none: bool = False, optional: bool = False, default: bool = None) -> (bool, None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional
        assert default is None or isinstance(default, bool), default

        if not self.src_.contains(name):
            if optional:
                return default
            raise self._exc('Parameter {} is missing'.format(name))

        v = self.src_.get_raw_value(name)
        if v is None:
            if allow_none:
                return None
            raise self._exc('Parameter {} is None'.format(name))

        if not isinstance(v, bool):
            raise self._exc('Parameter {}={} should be bool, not {}'.format(name, v, type(v)))

        return v

    def get_int(self, name: str, allow_none: bool = False, optional: bool = False, default: int = None) -> (str, None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional
        assert default is None or isinstance(default, int), default

        if not self.src_.contains(name):
            if optional:
                return default
            raise self._exc('Parameter {} is missing'.format(name))

        v = self.src_.get_raw_value(name)
        if v is None:
            if allow_none:
                return None
            raise self._exc('Parameter {} is None'.format(name))

        if not isinstance(v, int):
            raise self._exc('Parameter {}={} should be int, not {}'.format(name, v, type(v)))

        return v

    def get_dict(self, name: str, dic: dict, optional: bool = False, default=None):
        assert isinstance(name, str), name
        assert isinstance(dic, dict), dic
        assert isinstance(optional, bool), optional

        if not self.src_.contains(name):
            if optional:
                return default
            raise self._exc('Parameter {} is missing'.format(name))

        v = self.src_.get_raw_value(name)

        if v in dic:
            return dic[v]
        else:
            raise self._exc('Attribute "{}={}" is invalid. Available values: {}'.format(name, v, dic.keys()))


class DictReader(NamedParameters, INamedValues):
    def __init__(self, d: dict, exc_type=Warning, location: ILocation = None):
        assert isinstance(d, dict), d

        super().__init__(self, exc_type, location)

        self.d_ = d

    def contains(self, name: str) -> bool:
        return name in self.d_

    # None if does not contain
    def get_raw_value(self, name: str):
        return self.d_.get(name)



from ..lib.stddef import ValueType
from ..lib.variant import Variant


class StdValueCvt:      # pylint: disable=too-few-public-methods
    def try_convert(self, src_val: Variant, dest_val: Variant) -> bool:     # pylint: disable=no-self-use
        assert isinstance(src_val, Variant), src_val
        assert isinstance(dest_val, Variant), dest_val

        if src_val.is_void() or dest_val.is_void():
            return False

        if src_val.same_type(dest_val):
            dest_val.copy_from_same_type(src_val)
            return True

        if src_val.vt() == ValueType.VT_STR or dest_val.vt() == ValueType.VT_STR:
            return False

        src_t = src_val.vt()
        dest_t = dest_val.vt()
        src_v = src_val.get_value()
        dest_v = None

        if src_v is not None:
            if src_t == ValueType.VT_BOOL:

                # if dest_t == ValueType.VT_BOOL:
                if dest_t == ValueType.VT_INT:
                    dest_v = 1 if src_v else 0
                elif dest_t == ValueType.VT_FLOAT:
                    dest_v = 1.0 if src_v else 0.0
                else:
                    assert False, dest_t

            elif src_t == ValueType.VT_INT:

                if dest_t == ValueType.VT_BOOL:
                    dest_v = bool(src_v)
                # elif dest_t == ValueType.VT_INT:
                elif dest_t == ValueType.VT_FLOAT:
                    dest_v = float(src_v)
                else:
                    assert False, dest_t

            elif src_t == ValueType.VT_FLOAT:

                if dest_t == ValueType.VT_BOOL:
                    dest_v = bool(src_v)
                elif dest_t == ValueType.VT_INT:
                    dest_v = int(src_v)
                # elif dest_t == ValueType.VT_FLOAT:
                else:
                    assert False, dest_t

            else:
                assert False, src_t

        dest_val.set_value_w(dest_v)
        return True



#!/usr/bin/python3

import logging

from twisted.internet.protocol import Protocol, Factory

# Volcano
from ..lib.bin import bytes_to_str
from ..lib.modbus import MbException, MbLimits, MbFunctionCode, MbError

# locals
from .mb_srv_protocol import ModbusRequestReadWords, ModbusRequestReadBits, parse_req_tcp, build_response_exception, build_response_read_words, build_response_read_bits


MB_EXC_ILLEGAL_FN_CODE = 0x01
MB_EXC_GATEWAY_PATHS_NA = 0x0A


class DropProtocol(Protocol):
    def connectionMade(self):
        self.transport.loseConnection()


class MbConnection(Protocol):
    def __init__(self, factory, devices, env):
        self.factory_ = factory
        self.devices_ = devices
        self.rcv_buf_ = bytearray()
        self.log = logging.getLogger(env.name)

    def process_mber_mbex(self, req: (ModbusRequestReadWords, ModbusRequestReadBits)):
        slave_nb = req.slave_nb
        fn_nb = req.fn_nb

        is_read_bits = None
        if fn_nb in (MbFunctionCode.MB_FN_READ_OUT_BITS_1, MbFunctionCode.MB_FN_READ_IN_BITS_2):
            is_read_bits = True
        elif fn_nb in (MbFunctionCode.MB_FN_READ_OUT_WORDS_3, MbFunctionCode.MB_FN_READ_IN_WORDS_4):
            is_read_bits = False
        else:
            raise MbException(MB_EXC_ILLEGAL_FN_CODE, 'Modbus function {} is not supported'.format(fn_nb))
        
        # С версии 1.1.12 может быть несколько устройств с одним slaveNb - отвечает первое, в адресную зону которого попадает запрос
        found = False
        
        if is_read_bits:
            resp_bits = [False] * req.nb_bits
            for slave in self.devices_:
                if slave.slave_nb() == slave_nb:
                    found = True
                    slave.mb_read_bits_mbex(req.addr, req.nb_bits, resp_bits)  # can raise MbException/MbError
                    
            if found:
                return build_response_read_bits(req, resp_bits)
        else:
            resp_bytes = bytearray(req.nb_words * 2)
            
            for slave in self.devices_:
                if slave.slave_nb() == slave_nb:
                    found = True
                    slave.mb_read_words_mbex(req.addr, req.nb_words, resp_bytes)  # can raise MbException/MbError
                    
            if found:
                return build_response_read_words(req, resp_bytes)
        
        raise MbException(MB_EXC_GATEWAY_PATHS_NA, 'Slave {0} (0x{0:x}) does not exist'.format(slave_nb))


    def dataReceived(self, data: bytes):
        assert isinstance(data, bytes), data

        self.log.debug('R< %s', bytes_to_str(data))
        self.rcv_buf_ += data
        try:
            while self.rcv_buf_:

                # MbError, None, {}
                req = parse_req_tcp(self.rcv_buf_)  # throws
                if req is None:  # not enough data
                    if len(self.rcv_buf_) >= MbLimits.MAX_FRAME_SIZE_BYTES:
                        raise MbError('Modbus rcv buffer overflow')
                    else:
                        return

                assert isinstance(req, (ModbusRequestReadWords, ModbusRequestReadBits)), req
                assert 0 < req.size_bytes <= len(self.rcv_buf_), req

                self.rcv_buf_ = self.rcv_buf_[req.size_bytes:]

                try:
                    res = self.process_mber_mbex(req)
                    self.log.debug('S> %s', bytes_to_str(res))
                    self.transport.write(res)
                except MbException as ex:
                    self.log.warning(ex)
                    res_ba = build_response_exception(req, ex.code)
                    self.transport.write(res_ba)
        except MbError as e:
            self.log.warning(e)
            self.transport.loseConnection()

    def connectionMade(self):
        self.log.debug('Modbus client connected')

    def connectionLost(self, reason):   # pylint: disable=signature-differs
        self.log.debug('Modbus client disconnected')
        self.factory_.on_connection_lost(self)


class MbFactory(Factory):
    max_connections = None

    def __init__(self, devices, env):
        super().__init__()

        self.max_connections = env.max_con
        self.env = env
        self.devices_ = devices
        self.connections_ = []
        self.log = logging.getLogger(env.name)

    def buildProtocol(self, addr):
        if len(self.connections_) >= self.max_connections:
            self.log.warning('Max number of connections [%s] reached', self.max_connections)
            return DropProtocol()

        c = MbConnection(self, self.devices_, self.env)
        self.connections_.append(c)
        return c

    def on_connection_lost(self, con):
        self.connections_.remove(con)



#!/usr/bin/python3

import time


class Timer:
    def __init__(self):
        self.time_ = time.perf_counter()

    def set_now(self):
        self.time_ = time.perf_counter()

    def secs(self):
        rval = time.perf_counter() - self.time_
        return rval if rval >= 0.0 else 0.0


class Timeout:
    def __init__(self, timeout_sec=0.0, expired=False):
        self.timeout_sec_ = timeout_sec
        self.expired_ = expired
        if not self.expired_:
            self.t1_ = time.perf_counter()

    def is_expired(self):
        return self.remain_sec() == 0.0

    def remain_sec(self):
        if self.expired_:
            return 0.0
        passed_sec = time.perf_counter() - self.t1_
        if passed_sec > self.timeout_sec_:
            self.expired_ = True
            return 0.0
        return passed_sec

    def start(self, timeout_sec=None):
        if timeout_sec is not None:
            self.timeout_sec_ = timeout_sec
        self.expired_ = self.timeout_sec_ <= 0.0
        if not self.expired_:
            self.t1_ = time.perf_counter()



#!/usr/bin/python3

def search(it, cb):
    for x in it:
        if cb(x):
            return x
    return None


def unique_name(prefix, cb):
    ctr = 1
    name = '{}{}'.format(prefix, ctr)
    while not cb(name):
        ctr += 1
        name = '{}{}'.format(prefix, ctr)
    return name



import collections

from ..lib.bin import get_crc16_as_bytes, hi_byte, lo_byte
from ..lib.modbus import MbLimits, MbFunctionCode, ErrorCode, MbError

ModbusRequestReadWords = collections.namedtuple('ModbusRequestReadWords', 'size_bytes trans_id slave_nb fn_nb addr nb_words')
ModbusRequestReadBits = collections.namedtuple('ModbusRequestReadBits', 'size_bytes trans_id slave_nb fn_nb addr nb_bits')


def _word_ml(ba: (bytes, bytearray), offset: int) -> int:
    assert isinstance(ba, (bytes, bytearray)), ba
    assert isinstance(offset, int), offset
    assert len(ba) > (offset + 1)

    return (ba[offset] << 8) | ba[offset + 1]


# Response:
#   - None = not enough data
#   - MbError,  (MbException not raisen!)
#   - (bytes_parsed: int, request: MbServerRequest)
def parse_req_tcp(ba: (bytes, bytearray), offset: int = 0, nb: (None, int) = None) -> (None, ModbusRequestReadWords, ModbusRequestReadBits):
    assert isinstance(ba, (bytes, bytearray)), ba
    assert isinstance(offset, int), offset
    assert nb is None or isinstance(nb, int), nb

    if nb is None:
        nb = len(ba) - offset

    if nb < 6:
        return None

    rtu_size = _word_ml(ba, offset + 4)
    tcp_size = rtu_size + 6

    if nb < tcp_size:
        return None

    if rtu_size < 2:
        raise MbError(ErrorCode.BadData)

    rtu_offset = 6
    trans_id = _word_ml(ba, offset + 0)
    slave_nb = ba[rtu_offset + 0]
    fn_nb = ba[rtu_offset + 1]

    if fn_nb in (MbFunctionCode.MB_FN_READ_OUT_BITS_1, MbFunctionCode.MB_FN_READ_IN_BITS_2):
        # REQ: Slave,Fn,AddrH,AddrL,NbBitsH,NbBitsL
        if rtu_size != 6:
            raise MbError(ErrorCode.BadData)

        addr = _word_ml(ba, rtu_offset + 2)
        nb_bits = _word_ml(ba, rtu_offset + 4)

        if nb_bits == 0 or nb_bits > MbLimits.MAX_READ_BITS_NB or (addr + nb_bits - 1) > 0xffff:
            raise MbError(ErrorCode.BadData)

        return ModbusRequestReadBits(tcp_size, trans_id, slave_nb, fn_nb, addr, nb_bits)

    elif fn_nb in (MbFunctionCode.MB_FN_READ_OUT_WORDS_3, MbFunctionCode.MB_FN_READ_IN_WORDS_4):
        # REQ: Slave,Fn,AddrH,AddrL,NbWordsH,NbWordsL
        if rtu_size != 6:
            raise MbError(ErrorCode.BadData)

        addr = _word_ml(ba, rtu_offset + 2)
        nb_words = _word_ml(ba, rtu_offset + 4)

        if nb_words == 0 or nb_words > MbLimits.MAX_READ_WORDS_NB or (addr + nb_words - 1) > 0xffff:
            raise MbError(ErrorCode.BadData)

        return ModbusRequestReadWords(tcp_size, trans_id, slave_nb, fn_nb, addr, nb_words)

    else:
        raise MbError(ErrorCode.UnknownFn)


def _wrap_res_tcp(trans_id: int, rtu_data: (bytes, bytearray)) -> bytes:
    assert isinstance(trans_id, int), trans_id
    assert isinstance(rtu_data, (bytes, bytearray)), rtu_data

    if trans_id < 0 or trans_id > 0xffff:
        raise MbError(ErrorCode.BadData)

    return bytes([hi_byte(trans_id), lo_byte(trans_id), 0, 0, hi_byte(len(rtu_data)), lo_byte(len(rtu_data))]) + rtu_data


def _wrap_res_rtu(rtu_data: (bytes, bytearray)) -> bytes:
    assert isinstance(rtu_data, (bytes, bytearray)), rtu_data

    return bytes(rtu_data + get_crc16_as_bytes(rtu_data))


def build_response_exception(req: (ModbusRequestReadWords, ModbusRequestReadBits), exc_code: int) -> bytes:
    assert isinstance(req, (ModbusRequestReadWords, ModbusRequestReadBits)), req
    assert isinstance(exc_code, int), exc_code

    if req.slave_nb < 0 or req.slave_nb > 0xff:
        raise MbError(ErrorCode.BadData)

    if req.fn_nb < 0 or req.fn_nb > 0xff:
        raise MbError(ErrorCode.BadData)

    if exc_code <= 0 or exc_code > 0xff:
        raise MbError(ErrorCode.BadData)

    resp = bytes([req.slave_nb, req.fn_nb | 0x80, exc_code])
    return _wrap_res_rtu(resp) if req.trans_id is None else _wrap_res_tcp(req.trans_id, resp)


def build_response_read_words(req: ModbusRequestReadWords, data_bytes: (bytes, bytearray)) -> bytes:
    assert isinstance(req, ModbusRequestReadWords), req
    assert isinstance(data_bytes, (bytes, bytearray)), data_bytes
    assert len(data_bytes) == (req.nb_words * 2), data_bytes

    # REPLY: Slave,Fn,NbBytes,DATA
    rtu = bytes([req.slave_nb, req.fn_nb, len(data_bytes)]) + data_bytes

    return _wrap_res_rtu(rtu) if req.trans_id is None else _wrap_res_tcp(req.trans_id, rtu)


def build_response_read_bits(req: ModbusRequestReadBits, bits) -> bytes:
    assert isinstance(req, ModbusRequestReadBits), req
    assert bits
    assert len(bits) == req.nb_bits, (bits, req.nb_bits)

    # REPLY: Slave,Fn,NbBytes,DATA
    nb_bytes = (req.nb_bits + 7) // 8
    d_bytes = bytearray()
    src_bit_idx = 0
    for i_byte in range(nb_bytes):  # pylint: disable=unused-variable
        byte_val = 0
        for i_bit in range(8):
            if bits[src_bit_idx]:
                byte_val |= 1 << i_bit
            src_bit_idx += 1
            if src_bit_idx == req.nb_bits:
                break
        d_bytes.append(byte_val)

    rtu = bytes([req.slave_nb, req.fn_nb, len(d_bytes)]) + d_bytes

    return _wrap_res_rtu(rtu) if req.trans_id is None else _wrap_res_tcp(req.trans_id, rtu)



#!/usr/bin/python3
import argparse
import logging
import sys
from time import perf_counter

from twisted.internet.protocol import Factory, Protocol
from twisted.internet import reactor, stdio, task
from twisted.internet.error import CannotListenError

# volcano
from ..lib.stddef import VOLCANO_DEFAULT_TCP_PORT, Quality
from ..lib.log import configure_arg_parser_for_log, configure_logger
from ..lib.xml_reader import LoadException
from ..lib import variant

# locals
from .db import Db, Tag
from .cvt import StdValueCvt
from .cp import TTYControlPanel
from .connection import Connection


class DropProtocol(Protocol):
    def connectionMade(self):
        self.transport.loseConnection()


def configure_my_args(parser):
    parser.add_argument('-f', '--file', help='Config file', default='db.xml')
    parser.add_argument('-i', '--iface', help='Listen interface', default='')
    parser.add_argument('-p', '--port', help='Listen port', type=int, default=VOLCANO_DEFAULT_TCP_PORT)
    parser.add_argument('--demo', help='Turn on demo mode', action='store_true')
    parser.add_argument('--max_con', help='Max number of connections', type=int, default=5, choices=range(1, 10))

    parser.add_argument('--log_tag_upd', help='Log tag udates', action='store_true')


class VolcanoServer(Factory):
    def __init__(self, db, env, log):
        self.db = db        # pylint: disable=invalid-name
        self.env = env
        self.log = log

        self.connections = []
        self.value_cvt_ = StdValueCvt()

        self.sys_tags_ = {
            'fps_in': self.db.create_tag_by_long_name(None, 'sys.fps_in', variant.IntValue(), Quality.QUALITY_GOOD),
            'fps_out': self.db.create_tag_by_long_name(None, 'sys.fps_out', variant.IntValue(), Quality.QUALITY_GOOD),
        }
        self.nb_msg_in_ = 0
        self.nb_msg_out_ = 0
        self.fps_tstamp_ = perf_counter()

        self.timer_handler_ = task.LoopingCall(self.on_timer_)  # stop() will stop the looping calls
        self.timer_handler_.start(1.0)

    def value_cvt(self):
        return self.value_cvt_

    def buildProtocol(self, addr):
        if len(self.connections) >= self.env.max_con:
            self.log.warning('Max number of connections (%s pcs) reached, new one is dropped', self.env.max_con)
            return DropProtocol()

        con = Connection(self, self.db, self.env, self.log)

        self.connections.append(con)
        self.log.info('New connection. Nb connections: %s', len(self.connections))
        return con

    def forget(self, con):
        self.connections.remove(con)
        self.log.info('Connection removed. Nb connections: %s', len(self.connections))

    def on_timer_(self):
        t = perf_counter()
        secs = t - self.fps_tstamp_
        if secs > 0:
            fps_in = self.nb_msg_in_ / secs
            fps_out = self.nb_msg_out_ / secs
            self.sys_tags_['fps_in'].set_strict(int(fps_in), Quality.QUALITY_GOOD, None)
            self.sys_tags_['fps_out'].set_strict(int(fps_out), Quality.QUALITY_GOOD, None)

            self.notify_satelites_tag_changed(self.sys_tags_.values())

        self.nb_msg_in_ = 0
        self.nb_msg_out_ = 0
        self.fps_tstamp_ = t

    def notify_satelites_tag_changed(self, tag_or_tags: (Tag, list, tuple)) -> None:
        tags = (tag_or_tags,) if isinstance(tag_or_tags, Tag) else tag_or_tags

        for tag in tags:
            assert isinstance(tag, Tag), tag
            assert not tag.is_void(), tag

            for c in self.connections:
                c.notify_update(tag)


def main():
    try:
        # Args
        arg_parser = argparse.ArgumentParser()
        configure_my_args(arg_parser)
        configure_arg_parser_for_log(arg_parser)  # let logger add his arguments for later configure_logger() call
        env = arg_parser.parse_args()

        # Logging
        configure_logger(env)
        log = logging.getLogger('core')  # should log under our instance name

        # myself = 'volcano-core'
        # log.info('Volcano v%s', pkg_resources.get_distribution(myself).version)

        # Database
        db = Db(log)
        try:
            if env.demo:
                log.info('!! DEMO MODE !! Database is not loaded, creating demo structure')
                db.setup_demo_mode()
            else:
                db.upload_file(env.file)
                log.info('Loaded %s tags', db.nb_tags())
        except LoadException as ex:
            log.error(ex)
            return 1

        server = VolcanoServer(db, env, log)
        con_server = TTYControlPanel(server)
        stdio.StandardIO(con_server)

        log.info('Start listen: iface=%s port=%s max_con=%s', env.iface, env.port, env.max_con)
        try:
            reactor.listenTCP(env.port, server, interface=env.iface)    # pylint: disable=no-member
        except CannotListenError as ex:
            log.error(ex)
            return 1

        reactor.run()                                               # pylint: disable=no-member

        log.info('GOODBYE')

        return 0

    finally:
        logging.shutdown()


if __name__ == '__main__':
    sys.exit(main())



import datetime
import collections

from .num_ops import bcd_to_int, make_word, hi_byte, lo_byte
from .merk_nzif_common import ProtocolException, NackException


PROFILE_RECORD_SZ = 8
MEM_LEN = 65536

MeterInfo = collections.namedtuple('MeterInfo', 'Address PulseToPhysicalUnitsFactor MemorySize MaxNbOfAddrToReadAtOnce')

CrProfileInfo = collections.namedtuple('CrProfileInfo', 'Address Overflow DateTime')


class NzifProfiles:
    def __init__(self, mi: MeterInfo):
        self.meter_info_ = mi
        
        self.LastReadAddress = -1       # if not yet read
        self.HasActiveHeader = False    # true => after last read header we didnt finish all its data-records
        
        # Defined if HasActiveHeader == True:
        self.HeaderDtmUtc = None                    # Date of the last header
        self.NbDataRecordsParsedSinceHeader = None 
        self.AggIvlMins = None


    def set_header(self, dtm_utc: datetime.datetime, agg_ivl: int):
        self.HasActiveHeader = True
        self.HeaderDtmUtc = dtm_utc
        self.NbDataRecordsParsedSinceHeader = 0
        self.AggIvlMins = agg_ivl

    def read_cr_profile_info(self, ctx, log) -> CrProfileInfo:
        # 2.4.3.5
        req = bytes((self.meter_info_.Address, 0x08, 0x04))
        
        ctx.send_rcv(req, 7, 'ReadCrProfileInfo', log)

        overflow = (ctx.data(0) & 0x80) > 0

        try:
            m = bcd_to_int(ctx.data(0) & 0x7f)
            h = bcd_to_int(ctx.data(1))
            D = bcd_to_int(ctx.data(2))
            M = bcd_to_int(ctx.data(3))
            Y = bcd_to_int(ctx.data(4)) + 2000
            
            dtm = datetime.datetime(Y, M, D, h, m)
        except ValueError as ex:
            log.warn(ex)
            raise ProtocolException(str(ex))

        cr_profile_addr = make_word(ctx.data(5), ctx.data(6))

        log.debug('Current profile: %s; Overflow: %s; Address: %s', dtm, overflow, cr_profile_addr)

        if (cr_profile_addr % 8) != 0:
            raise ProtocolException('Current profile address is not modulous to 8: {}'.format(cr_profile_addr))

        return CrProfileInfo(cr_profile_addr, overflow, dtm)

    # !NackException (subclass of ProtocolException)
    def read_once(self, ctx, start_addr: int, nb_recs: int, cr_profile_info: CrProfileInfo, log):
    
        log.debug('Read %s records from addr %s', nb_recs, start_addr)

        data_sz = nb_recs * PROFILE_RECORD_SZ

        # par 2.4.4
        req = bytes((self.meter_info_.Address, 0x06, 0x03, hi_byte(start_addr), lo_byte(start_addr), data_sz))
        
        try:
            ctx.send_rcv(req, data_sz, 'ReadProfileMemory', log)
        except NackException as ex:
            # 2.4.4.1
            # Если в конце физической памяти массива профиля отсутствует место для
            # записи целого часа (заголовка часа и часовых срезов мощности), то эта область 
            # памяти не используется, не очищается и может содержать любую информацию. Ее нельзя читать.
            # Заголовок следующего часа начинается с нулевого адреса и устанавливается признак переполнения массива профиля мощности.
            # По факту это значит, что попытка читать зону в конце часто оборачивается ответом NACK<1>
            if ex.nack_code == 1:
                # действительно мы в самом конце? попробуем считывать по 1 записи
                next_addr = start_addr + data_sz
                if next_addr < 65400:
                    return

                if nb_recs > 1:
                    log.debug('Reached end of memory, start reading by 1')
                    self.read_once(ctx, start_addr, 1, cr_profile_info, log)
                else:
                    log.debug('Reached end of memory, goto 0')
                    self.HasActiveHeader = False
                    self.LastReadAddress = MEM_LEN - PROFILE_RECORD_SZ
                    return
            else:
                raise

        self.LastReadAddress = start_addr + data_sz - PROFILE_RECORD_SZ

        for rec in range(nb_recs):
            offset = rec * PROFILE_RECORD_SZ
            rec_addr = start_addr + offset

            if not self.HasActiveHeader:
                try:
                    tpl = _parse_rec_header(ctx.data(), offset, log)
                except Exception as ex:
                    log.warn('Rec %s: header is invalid, skipping', rec_addr)
                    continue
                
                if not tpl:
                    # Поскольку мы начинаем наше чтение с произвольной записи (особенно при переполненном буфере), 
                    # то мы можем наткнуться на незаголовок. Просто пропускаем до следующего заголовка
                    log.debug('Rec %s: header not set, skipping', rec_addr)
                    continue
                
                aggIvl, dtmUTC = tpl

                nb_data_recs_after_hdr = 60 / aggIvl

                last_data_rec_addr_of_cr_hour = rec_addr + nb_data_recs_after_hdr * PROFILE_RECORD_SZ

                if last_data_rec_addr_of_cr_hour >= self.meter_info_.MemorySize:
                    # warn - потому что в этом случае не должно быть HEADER-записи
                    log.warn('Profile header doesnt fit into an hour')
                    self.HasActiveHeader = False
                    # чтобы на следующем цикле считать адрес 0
                    self.LastReadAddress = self.meter_info_.MemorySize - PROFILE_RECORD_SZ
                    return

                if rec_addr < cr_profile_info.Address and last_data_rec_addr_of_cr_hour >= cr_profile_info.Address:
                    # не умещается до текущего указателя
                    log.debug('Finished reading whole buffer')
                    self.HasActiveHeader = False
                    self.LastReadAddress = -1
                    return

                # LOG_DEBUG(lg, "Rec " << rec_addr << ": HEADER: " << dtmUTC << "(utc), ivl " << aggIvl << " min")
                self.set_header(dtmUTC, aggIvl)
                
            else:
                minutes_add = self.AggIvlMins * self.NbDataRecordsParsedSinceHeader

                self.NbDataRecordsParsedSinceHeader += 1

                pulse_pp, pulse_pn, pulse_qp, pulse_qn = _parse_data_rec(ctx.data(), offset)

                log.debug('Rec %s: PP=%s; PN=%s; QP=%s; QN=%s', rec_addr, pulse_pp, pulse_pn, pulse_qp, pulse_qn)

                if (minutes_add + self.AggIvlMins) >= 60:
                    log.debug('Finished hour')
                    self.HasActiveHeader = False


    # ProtocolException
    def read_profiles(self, ctx, max_read_requests, log):
        cr_profile = self.read_cr_profile_info(ctx, log)

        # page 150
        for _i in range(max_read_requests):
            # Check stop?
            
            frame = self.select_next_frame_(None if self.LastReadAddress == -1 else self.LastReadAddress, cr_profile, log)

            if not frame:
                # если lastReadAddr==-1, это значит, что мы ничего пока не читали и счетчик сейчас ведет только самый 1й профиль
                if self.LastReadAddress != -1:
                    log.info('Finished reading whole buffer')

                self.HasActiveHeader = False
                self.LastReadAddress = -1
                
                return
            
            next_addr, nb_recs = frame

            self.read_once(ctx, next_addr, nb_recs, cr_profile, log)

    # last_read_addr: None если чтений не было, иначе последний прочтенный адрес
    
    # None => no need to read
    # (next_addr, nb_recs)
    def select_next_frame_(self, last_read_addr: (int, None), cr_profile_info: CrProfileInfo, _log) -> bool:
        # 2.4.4.1
        MAX_NB_RECORDS_IN_REQUEST = self.meter_info_.MaxNbOfAddrToReadAtOnce // PROFILE_RECORD_SZ

        # Если признак =0, то массив профиля мощности не переполнен
        # и его первая запись (от момента инициализации массива) находится по физическому адресу
        # 0000h. Если признак =1, то массив срезов переполнен, и новые записи пишутся поверх самых
        # старых. При этом, первая запись (самая старая) находится по адресу указателя.
        if last_read_addr is None:  # новый цикл чтения
            if cr_profile_info.Overflow:
                # Поскольку есть вероятность, что за время, пока мы готовимся считать,
                # указатель перейдет на следующую запись, то мы читаем не самую старую, а с небольшим отступом
                next_addr = (cr_profile_info.Address + 5 * PROFILE_RECORD_SZ) % MEM_LEN

                if next_addr < cr_profile_info.Address:  # новый цикл чтения (переход через конец памяти)
                    nb_recs = MAX_NB_RECORDS_IN_REQUEST
                else:
                    absmax = (MEM_LEN - next_addr) // PROFILE_RECORD_SZ
                    nb_recs = min(MAX_NB_RECORDS_IN_REQUEST, absmax)
            else:
                next_addr = 0
                nb_recs = min(MAX_NB_RECORDS_IN_REQUEST, cr_profile_info.Address // PROFILE_RECORD_SZ)

                if nb_recs == 0:
                    return None
                    
            return (next_addr, nb_recs)
        else:
            # следующая запись за последней считанной
            next_addr = (last_read_addr + PROFILE_RECORD_SZ) % MEM_LEN

            # уперлись в текущий профиль
            if next_addr == cr_profile_info.Address:
                return None

            if next_addr < cr_profile_info.Address: 
                absmax = (cr_profile_info.Address - next_addr) // PROFILE_RECORD_SZ
            else:
                absmax = (MEM_LEN - next_addr) // PROFILE_RECORD_SZ

            return (next_addr, min(MAX_NB_RECORDS_IN_REQUEST, absmax))



# returns tuple (Address: int, Overflow: bool, DateTime: datetime)

# None -> это не header (ф-ция может использоваться для поиска хедера)
# (agg_ivl [minutes]: int, dtm: datetime)
# !ProtocolException  -  это хедер, но в нем ошибка (например, неверный BCD код или плохая дата)
def _parse_rec_header(d, offset: int, log) -> tuple:
    # По опыту у header-записей 8й байт равен FF или 0, но из документации
    # следует, что он вообще игнорируется,поэтому приходится полагаться только на CRC

    # check CRC
    _sum = d[offset+0] + d[offset+1] + d[offset+2] + d[offset+3] + d[offset+4] + d[offset+5]
    if d[offset+6] != (_sum & 0xff):
        # Это не ошибка, ведь функция может использоваться и для поиска заголовка
        return None

    try:
        h = bcd_to_int(d[offset+0])
        D = bcd_to_int(d[offset+1])
        M = bcd_to_int(d[offset+2])
        Y = bcd_to_int(d[offset+3])
        
        dtm_meter_tz = datetime.datetime(Y, M, D, h, 0)
    except ValueError as ex:
        log.warn(ex)
        #con_val.invalidate(Quality.QUALITY_NA)
        raise ProtocolException(str(ex))
    
    # !!! Meter TZ can be not a local time
    # *out_DtmUTC = SE::Platform::DateTime::localToUtc( &dtmMeterTZ );

    agg_ivl = d[offset+5]

    if agg_ivl == 0 or agg_ivl > 60 or (60 % agg_ivl) != 0:
        raise ProtocolException('Profile header contains invalid agg ivl {} (minutes)'.format(agg_ivl))

    return (agg_ivl, dtm_meter_tz)

# (pp|None, pn|None, qp|None, qn|None)
def _parse_data_rec(d, offset: int) -> tuple:
    # стр 101 [2.4.3.6.1] - сообщает о маскировании
    pp_good = (d[offset+0] & 0x80) == 0
    pn_good = (d[offset+2] & 0x80) == 0
    qp_good = (d[offset+4] & 0x80) == 0
    qn_good = (d[offset+6] & 0x80) == 0

    pp = make_word(d[offset + 0] & 0x7f, d[offset + 1]) if pp_good else None
    pn = make_word(d[offset + 2] & 0x7f, d[offset + 3]) if pn_good else None
    qp = make_word(d[offset + 4] & 0x7f, d[offset + 5]) if qp_good else None
    qn = make_word(d[offset + 6] & 0x7f, d[offset + 7]) if qn_good else None

    return (pp, pn, qp, qn)



import logging

_LEVELS = {
    'all': logging.NOTSET,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warn': logging.WARNING,
    'error': logging.ERROR,
}


def configure_arg_parser_for_log(args_parser) -> None:
    args_parser.add_argument('-l', '--log', help='Log level', default='info', choices=['all', 'debug', 'info', 'warn', 'error', 'none'])


def configure_logger(env) -> None:
    assert env.log, env

    msg_fmt = '%(levelname)s %(asctime)s [%(name)s] %(message)s'

    if env.log == 'none':
        logging.disable(logging.CRITICAL)
    else:
        lev = _LEVELS.get(env.log, None)
        if lev is None:
            raise SystemExit('Invalid value for logging level: "{}". Use all,debug,info,warn,error.'.format(env.log))
        logging.basicConfig(level=lev, format=msg_fmt)

    logging.addLevelName(logging.CRITICAL, '!!!')
    logging.addLevelName(logging.ERROR, '!! ')
    logging.addLevelName(logging.WARNING, '!  ')
    logging.addLevelName(logging.INFO, '   ')
    logging.addLevelName(logging.DEBUG, '   ')



#!/usr/bin/python3
# encoding=utf-8

from ..lib.xml_reader import XmlReader


class BaseDevice:
    def __init__(self, my_xml_node, proposed_aux_name, parent_log):
        p = XmlReader(my_xml_node)

        self.branch_ = p.get_str('branch')
        self.aux_name_ = p.get_str('name', proposed_aux_name)
        self.log_ = parent_log.getChild(self.aux_name_)
        self.go_offline_timeout_sec_ = p.get_time_secs('offlineTimeout', 5.0, min_val=0.0, max_val=600.0)

    def sync(self, services: dict):     # pylint: disable=no-self-use
        assert isinstance(services, dict), services

    def channel_closed_timeout(self):
        return self.go_offline_timeout_sec_

    def on_channel_closed(self):
        raise NotImplementedError()

    def branch(self, tag_name: str = None):
        return '{}.{}'.format(self.branch_, tag_name) if tag_name else self.branch_

    def aux_name(self):
        return self.aux_name_

    def log(self):
        return self.log_

    # не предполагает никаких exceptions
    # в случае проблем может просто закрыть канал
    def work(self, ctx) -> None:
        raise NotImplementedError()

    def get_tag_full_name(self, tag_short_name: str):
        # '@xxx' - global
        # 'xxx' - local
        if tag_short_name.startswith('@'):
            return tag_short_name[1:]
        else:
            return self.branch_ + '.' + tag_short_name



from twisted.protocols.basic import LineReceiver
from twisted.internet import reactor

from ..lib import tstamp


class TTYControlPanel(LineReceiver):
    delimiter = b'\n'

    def __init__(self, server):
        super().__init__()
        self.server = server
        self.last_cmd_ = None
        self.cwd_ = None

    def rawDataReceived(self, data):    # to prevent warnings on not implemented
        raise NotImplementedError()

    def lineReceived(self, line: bytes):
        s = line.decode('utf-8').strip()
        if s:
            self.last_cmd_ = s
            self.on_cmd(s)
        else:
            if self.last_cmd_:
                print(self.last_cmd_)
                self.on_cmd(self.last_cmd_)
            else:
                pass

    def on_cmd(self, cmd_str):
        items = cmd_str.split(' ')
        if not items:
            return

        try:
            cmd = items[0]
            if cmd == 'cd':
                self.cmd_cd_w(items)
            elif cmd == 'ls':
                self.cmd_ls_w(items)
            elif cmd == 'set':
                self.cmd_set_w(items)
            elif cmd == 'exit':
                reactor.stop()  # pylint: disable=no-member
            else:
                print('Unknown command: {}'.format(cmd))
        except Warning as ex:
            print(ex)

    def cmd_cd_w(self, items):

        def prn_tag(tag):
            if tag.has_children():
                if tag.is_void():
                    print('  > {:10}'.format(tag.short_name()))
                else:
                    print('  > {:10} {} {}'.format(tag.short_name(), tag.var().get_value(), tag.var().type_str()))
            else:
                if tag.is_void():
                    print('  - {:10}'.format(tag.short_name()))
                else:
                    print('  - {:10} {} {}'.format(tag.short_name(), tag.var().get_value(), tag.var().type_str()))

        if len(items) == 1:  # go to root
            self.cwd_ = None
            for child in self.server.db.enum_root_tags():
                prn_tag(child)

        elif len(items) == 2:
            tag_name = items[1]
            tag = self.server.db.find_tag_by_name(
                self.cwd_.full_name() + '.' + tag_name) if self.cwd_ else self.server.db.find_tag_by_name(tag_name)
            if tag:
                self.cwd_ = tag
                for child in tag.enum_child_tags():
                    prn_tag(child)
            else:
                raise Warning('Tag not found: {}'.format(tag_name))
        else:
            raise Warning('Expected 0 or 1 parameter')

    def cmd_ls_w(self, items):
        if len(items) == 1:
            for child in self.server.db.enum_root_tags():
                self.print_tag(child)

        elif len(items) == 2:
            tag_name = items[1]
            tag = self.server.db.find_tag_by_name(tag_name)
            if tag:
                for child in tag.enum_child_tags():
                    self.print_tag(child)
            else:
                raise Warning('Tag not found: {}'.format(tag_name))
        else:
            raise Warning('Expected 0 or 1 parameter')

    def cmd_set_w(self, items):
        if len(items) != 3:
            raise Warning('Expected 2 parameters: set TAG_NAME VALUE')

        tag_name = items[1]
        value_str = items[2]

        tag = self.server.db.find_tag_by_name(tag_name)
        if not tag:
            raise Warning('Tag not found: {}'.format(tag_name))

        if tag.is_void():
            raise Warning('Tag is VOID: {}'.format(tag_name))

        tag.var().parse_value_w(value_str)
        tag.quality = 0
        tag.tstamp_utc = tstamp.get_current_tstamp_utc()

        for c in self.server.connections:
            c.notify_update(tag)

    def print_tag(self, tag):   # pylint: disable=no-self-use
        if tag.is_void():
            print('  - {:10}'.format(tag.short_name()))
        else:
            print('  - {:10} {} {}'.format(tag.short_name(), tag.var().get_value(), tag.var().type_str()))




import enum


VOLCANO_DEFAULT_TCP_PORT = 8091
VOLCANO_PROTOCOL_VERSION_STR = '0.2'


class ValueType(enum.Enum):
    VT_VOID = 'void'
    VT_BOOL = 'bool'
    VT_INT = 'int'
    VT_FLOAT = 'float'
    VT_STR = 'str'

    VT_ALL_LIST = (VT_VOID, VT_BOOL, VT_INT, VT_FLOAT, VT_STR)

    # Используется для сериализации в строку. Предполагается, что в дальнейшем могут
    # быть использованы не только строки для типов - в коде не должно быть жесткой привязки
    def stringify(self) -> str:
        return self.value

    # !ValueError
    @staticmethod
    def parse(s: str) -> 'ValueType':
        return ValueType(s)


class Quality(enum.IntEnum):    # IntFlag appeared in 3.6; Debian stretch has Py 3.5.3
    QUALITY_GOOD = 0
    QUALITY_COMM = 1
    QUALITY_NOT_INIT = 2
    QUALITY_NA = 4
    QUALITY_NA_TEMP = 8



import os.path
import json
# import copy
# from jsonschema import validate

from ..lib.named_parameters import DictReader

_DEFS = None
_NACK_CODES = {
    1: 'Недопустимая команда или параметр',
    2: 'Внутренняя ошибка счетчика',
    3: 'Не достаточен уровень доступа для удовлетворения запроса',
    4: 'Внутренние часы счетчика уже корректировались в течение текущих суток',
    5: 'Не открыт канал связи',
    6: 'Повторить запрос в течении 0,5 с',
    7: 'Не готов результат измерения по запрашиваемому параметру',
    8: 'Счетчик занят аналогичным процессом',
    9: 'Переполнение регистра единиц оплаты',
    10: 'Не возможно выполнить запрос управления',
}

'''
    type_code:  Код, возвращаемый счетчиком при отправке запроса на чтение Варианта Исполнения (2.4.3.24)
    ci:         Используется в расчетах измерений (см. 2.4.3.23; стр 116)
    ks:         Используется в расчетах измерений PQS (см. 2.4.3.23; стр 116)
    max_words_per_req:  2.4.4.1
                        В ответ на корректный запрос счетчик возвращает в поле данных ответа число байт, указанное
                        в поле «Число считываемых байт» запроса. Размер считываемого массива может составлять:
                        от 2 до 16 байт для счетчиков всех типов;
                        от 2 до 93 байт для счетчиков СЭТ-4ТМ.03, ПСЧ-4ТМ.05, ПСЧ-3ТМ.05;
                        от 2 до 133 байт для счетчиков ПСЧ-4ТМ.05М, ПСЧ-3ТМ.05М
    st_bytes:   Кол-во байт статуса. Для СЭТ 4ТМ 01 М ответ будет 4 байта, для остальных - 5
'''
def _load_type_defs(log) -> dict:

    if os.path.isfile('./nzif.json'):
        log.info('Using LOCAL file nzif.json')
        return _load_type_defs_from_file('./nzif.json')
    
    file_name = os.path.dirname(os.path.abspath(__file__)) + '/nzif.json'
    
    log.warn('Local file nzif.json not found, using: %s', file_name)
    
    return _load_type_defs_from_file(file_name)
    

def _load_type_defs_from_file(file_name: str) -> dict:
    assert isinstance(file_name, str), file_name

    try:
        with open(file_name) as f:
            data = json.loads(f.read())

        all_defs = {}  # includes both abstract and non-anstract

        # json является списком, а не словарем, потому что нужно упорядочение типов, чего нет в ключах словаря.
        # порядок требуется, потому что мы используем based_on
        for item in data:
            dr = DictReader(item, exc_type=Warning, location=None)  # explicitly say we dont need location from loader

            type_id = dr.get_str('id')
            if type_id in all_defs:
                raise Warning('Duplicated id "{}"'.format(type_id))

            based = item.get('based_on')
            if based:
                if isinstance(based, str):
                    if based not in all_defs:
                        raise Warning('Base type "{}" not found. Existing types: {}'.format(based, all_defs.keys()))
                    based_on_list = [all_defs[based]]
                else:
                    if not isinstance(based, list):
                        raise Warning('"based_on" should be str or list. not {}'.format(based))
                    based_on_list = []
                    for x in based:
                        if x not in all_defs:
                            raise Warning('Base type "{}" not found. Existing types: {}'.format(x, all_defs.keys()))
                        based_on_list.append(all_defs[x])
            else:
                based_on_list = []

            def base(ls, prop_name, default=None):
                for x in ls:
                    if prop_name in x:
                        return x[prop_name]
                return default

            new_type_def = {
                "id": type_id,
                "abstract": dr.get_bool('abstract', optional=True, default=False),
                "Inom": dr.get_int('Inom', optional=True, default=base(based_on_list, 'Inom')),
                "Unom": dr.get_str('Unom', optional=True, default=base(based_on_list, 'Unom')),
                "direct": dr.get_bool('direct', optional=True, default=base(based_on_list, 'direct')),
                "3_phase": dr.get_bool('3_phase', optional=True, default=base(based_on_list, '3_phase')),
                "type_code": dr.get_int('type_code', optional=True, default=base(based_on_list, 'type_code')),
                "profiles": dr.get_bool('profiles', optional=True, default=base(based_on_list, 'profiles')),
                "ci": dr.get_int('ci', optional=True, default=base(based_on_list, 'ci')),
                "ks": dr.get_int('ks', optional=True, default=base(based_on_list, 'ks')),
                "max_words_per_req": dr.get_int('max_words_per_req', optional=True, default=base(based_on_list, 'max_words_per_req')),
                "st_bytes": dr.get_int('st_bytes', optional=True, default=base(based_on_list, 'st_bytes')),
                'msr-1ph': [],
                'msr-3ph': [],
                'energy': [],
            }
            for bs in based_on_list:
                new_type_def['msr-1ph'].extend(bs.get('msr-1ph', []))
                new_type_def['msr-3ph'].extend(bs.get('msr-3ph', []))
                new_type_def['energy'].extend(bs.get('energy', []))

            # 1 phase
            # {"tag": "F", "f": "F_Hz", "rwri": [4, 0, 0]}
            for x in item.get('msr-1ph', []):
                if not isinstance(x, dict):
                    raise Warning('"msr-1ph" should contain dictionary items <tag: str, f: str, rwri: [...]>, not {}'.format(x))
                _dr = DictReader(x, exc_type=Warning)
                _dr.get_str('tag')
                _dr.get_str('f', options=('Ki', 'Ku', 'Ca', 'I_A', 'U_V', 'E_kwth', 'PQS_kwt', 'F_Hz', 'Cos'))
                rwri = _dr.get_raw('rwri')
                rwri_ok = isinstance(rwri, list) and len(rwri) == 3 and isinstance(rwri[0], int) and (0 <= rwri[0] <= 0xff) and isinstance(
                    rwri[1], int) and (0 <= rwri[1] <= 0xff) and isinstance(rwri[2], int) and (0 <= rwri[2] <= 0xff)
                if not rwri_ok:
                    raise Warning('"rwri" should be [p1: byte, p2: byte, p3: byte], not {}'.format(rwri))

                new_type_def['msr-1ph'].append(x)

            # 3 phase
            # {"tags": [null,  "Ia",   "Ib",   "Ic"],   "f": "I_A",     "rwri": [2, 0, null]},
            for x in item.get('msr-3ph', []):
                if not isinstance(x, dict):
                    raise Warning('"msr-3ph" should contain dictionary items <tags:[...], f: str, rwri: [...]>, not {}'.format(x))
                _dr = DictReader(x, exc_type=Warning)
                tags = _dr.get_raw('tags')
                tags_ok = isinstance(tags, list) and len(tags) == 4 and (tags[0] is None or isinstance(tags[0], str)) and isinstance(
                    tags[1], str) and isinstance(tags[2], str) and isinstance(tags[3], str)
                if not tags_ok:
                    raise Warning('"tags" should be [All: str | null, Ph1: str, Ph2: str, Ph3: str], not {}'.format(tags))

                _dr.get_str('f', options=('Ki', 'Ku', 'Ca', 'I_A', 'U_V', 'E_kwth', 'PQS_kwt', 'F_Hz', 'Cos'))
                rwri = _dr.get_raw('rwri')
                rwri_ok = isinstance(rwri, list) and len(rwri) == 3 and isinstance(rwri[0], int) and (0 <= rwri[0] <= 0xff) and isinstance(
                    rwri[1], int) and (0 <= rwri[1] <= 0xff) and rwri[2] in ('P', 'Q', None)
                if not rwri_ok:
                    raise Warning('"rwri" should be [p1: byte, p2: byte, p3: "P" | "Q" | null], not {}'.format(rwri))

                new_type_def['msr-3ph'].append(x)

            # Energy
            # {"tags": ["Eap", "Ean", "Erp", "Ern"], "arr": 0}
            for x in item.get('energy', []):
                if not isinstance(x, dict):
                    raise Warning('"energy" should contain dictionary items <tags:[...], arr: int>, not {}'.format(x))
                _dr = DictReader(x, exc_type=Warning)
                tags = _dr.get_raw('tags')
                tags_ok = isinstance(tags, list) and len(tags) == 4 and isinstance(tags[0], str) and isinstance(
                    tags[1], str) and isinstance(tags[2], str) and isinstance(tags[3], str)
                if not tags_ok:
                    raise Warning('"tags" should be [str, str, str, str], not {}'.format(tags))

                _dr.get_int('arr')

                new_type_def['energy'].append(x)

            if not new_type_def['abstract']:
                for n in ('Inom', 'Unom', 'direct', '3_phase', 'type_code', 'profiles', 'ci', 'ks', 'max_words_per_req', 'st_bytes'):
                    if new_type_def[n] is None:
                        raise Exception('{}: Type {} is not marked as abstract, but field "{}" is not set'.format(file_name, type_id, n))

            all_defs[type_id] = new_type_def

        return all_defs

    except Exception as ex:
        raise Warning('{}: {}'.format(file_name, ex))


def get_nzif_defs(log) -> dict:
    global _DEFS    # pylint: disable=global-statement

    if _DEFS is None:
        _DEFS = {}
        all_defs = _load_type_defs(log)
        for k, v in all_defs.items():
            if not v['abstract']:
                _DEFS[k] = v

    return _DEFS


def get_nzif_nack_codes() -> dict:
    global _NACK_CODES      # pylint: disable=global-statement
    return _NACK_CODES



from . import mb_ctx
from . import mb_dev
from . import base_prot
from . import tools


class MbProtocol(base_prot.BaseProtocol):
    def __init__(self, my_xml_node, comm_channel, proposed_name, parent_log):
        super().__init__(my_xml_node, comm_channel, mb_ctx.MbCtx(comm_channel), proposed_name, parent_log)

        for node in my_xml_node.findall('Device'):
            dev = mb_dev.MbDevice(node, tools.unique_name('device_', lambda x: not self.find_device(x)), self.log())
            if not dev.empty():
                self.add_device(dev)



#!/usr/bin/python3
import json
import json.decoder
from time import perf_counter

from twisted.internet.protocol import connectionDone
from twisted.protocols.basic import LineReceiver

# volcano
from ..lib import tstamp
from ..lib import variant

# locals
from .errors import ServerError, ServerWarning, ServerErrorCode
from .msg_parser import MsgParser

PROTOCOL_VERSION = (0, 2, 0)


class Connection(LineReceiver):
    delimiter = b'\n'

    def __init__(self, server, db, env, parent_log):
        self.server = server
        self.db = db            # pylint: disable=invalid-name
        self.env = env
        self.parent_log = parent_log

        self.inst_name_ = None      # None == not handshaked yet
        self.create_tm_ = perf_counter()
        self.log = self.parent_log.getChild('noname')

        self.sub_all_ = None    # (send_tstamp, use_id, , has_ch_uid, ch_uid)
        self.subs_map_ = None   # key=tag_id; value=(send_tstamp, use_id, , has_ch_uid, ch_uid)

    def connectionLost(self, reason=connectionDone):
        self.log.info('Lost connection')
        self.server.forget(self)

    def rawDataReceived(self, data):    # to prevent warnings on not implemented
        raise NotImplementedError()

    def lineReceived(self, line: bytes) -> None:
        assert isinstance(line, bytes), line

        d = line.strip(b' \t\r\n')
        if not d:
            return

        try:
            line_str = d.decode('utf-8', 'strict')
        except UnicodeDecodeError as ex:
            self.log.error('Error decoding command to utf-8: %s. Command: %s', ex, d)
            self.send({'c': 'drop', 'reason': ServerErrorCode.SyntaxError.value, 'comment': str(ex)})
            self.transport.loseConnection()
            return

        self.log.debug('R< ' + line_str)

        try:
            msg = json.loads(line_str)
        except json.decoder.JSONDecodeError as ex:
            self.log.error('%s. Command: %s. Connection will be dropped', ex, line_str)
            self.send({'c': 'drop', 'reason': ServerErrorCode.SyntaxError.value, 'comment': str(ex)})
            self.transport.loseConnection()
            return

        assert isinstance(msg, dict), msg

        try:
            self.process_cmd(msg)
        except ServerError as ex:
            self.log.error('%s. Command: %s. Connection will be dropped', ex, line_str)
            self.send({'c': 'drop', 'reason': ex.code.value, 'comment': str(ex)})
            self.transport.loseConnection()
        except ServerWarning as ex:
            self.log.warning('%s. Command: %s', ex, line_str)

    def process_cmd(self, msg: dict) -> None:
        dr = MsgParser(msg)
        cmd = dr.get_str('c')

        if cmd == 'hello':
            self.on_handshake(msg)
        else:
            if not self.inst_name_:
                raise ServerError('Command not allowed before handshake: {}'.format(msg), ServerErrorCode.ProtocolError)

            if cmd == 'sub':
                self.on_sub(msg)
            elif cmd == 'set':
                self.on_set(msg)
            elif cmd == 'find':
                self.on_find(msg)
            elif cmd == 'echo':
                self.send(msg)
            elif cmd == 'first':
                self.on_first(msg)
            elif cmd == 'next':
                self.on_next(msg)
            else:
                raise ServerError('Unknown message type "{}" in message {}'.format(cmd, msg), ServerErrorCode.ProtocolError)

    def on_handshake(self, msg: dict) -> None:
        if self.inst_name_:
            raise ServerError('Duplicated handshake message', ServerErrorCode.ProtocolError)

        dr = MsgParser(msg)

        inst_name = dr.get_str('name')
        proto_ver_s = dr.get_str('protocol')

        proto_ver = proto_ver_s.split('.')
        if len(proto_ver) not in (2, 3):
            raise ServerError('Invalid protocol version {}. Should be x.y or x.y.z'.format(proto_ver_s), ServerErrorCode.ProtocolError)

        try:
            clt_major = int(proto_ver[0])
            clt_minor = int(proto_ver[1])
        except ValueError:
            raise ServerError('Invalid protocol version {}. Should be x.y or x.y.z'.format(proto_ver_s), ServerErrorCode.ProtocolError)

        if clt_major != PROTOCOL_VERSION[0] or clt_minor > PROTOCOL_VERSION[1]:
            raise ServerError('Protocol version mismatch. Server: {}.{}.{}, client: {}'.format(PROTOCOL_VERSION[0], PROTOCOL_VERSION[1], PROTOCOL_VERSION[2], proto_ver_s), ServerErrorCode.BadVersion)

        self.inst_name_ = inst_name
        self.log = self.parent_log.getChild(self.inst_name_)
        self.log.debug('Connection renamed to "%s"', self.inst_name_)

        resp = {'c': 'welcome'}
        if 'ch' in msg:     # канал нужен в hs, потому что hs реализуется через сервисы, а они все работают на каналах
            resp['ch'] = msg['ch']

        self.send(resp)

    def on_set(self, msg):
        dr = MsgParser(msg)

        tag_name_or_id = dr.get_tag('tag')
        val_raw = dr.get_raw('v', allow_none=True)
        quality = dr.get_int('q', allow_none=False, optional=True, default=0)
        ts = dr.get_datetime('t', optional=True) or tstamp.get_current_tstamp_utc()

        tag = self.db.find_tag_by_name_or_id(tag_name_or_id)
        if not tag:
            raise ServerWarning('Cant set: tag {} doesnt exist'.format(tag_name_or_id), 'tag_not_found')

        if tag.is_void():
            raise ServerWarning('Cant set: tag {} is VOID'.format(tag.full_name()), 'tag_is_void')

        if tag.expr:
            raise ServerWarning('Cant set: tag {} is calculated'.format(tag.full_name()), 'tag_is_const')

        if val_raw is None:
            if tag.var().get_value() is None and tag.quality == quality:
                return
            tag.var().set_value_w(None)
        else:
            var = variant.Variant.try_from_raw_value(val_raw)
            if var is None:
                raise ServerError('Cant set tag {}={}: unable to detect value type'.format(tag_name_or_id, val_raw), ServerErrorCode.ProtocolError)

            tag_var_copy = tag.var().clone()
            if not self.server.value_cvt().try_convert(var, tag_var_copy):
                raise ServerWarning('Cant set "{}={}": value type mismatch'.format(tag.full_name(), val_raw), 'tag_value_type_mismatch')

            if tag_var_copy == tag.var() and tag.quality == quality:
                return

            tag.var().copy_from_same_type(tag_var_copy)

        tag.quality = quality
        tag.tstamp_utc = ts

        if self.env.log_tag_upd:
            self.log.debug('Set {}={}'.format(tag.full_name(), tag.var()))

        for c in self.server.connections:
            c.notify_update(tag)

        for dep in tag.expr_deps:
            assert dep.expr
            v, q = dep.expr.exec_safe(self.log)
            if not dep.var().val_equal(v, none_eq_none=True) or dep.quality != q:
                dep.var().set_value_w(v)
                dep.quality = q
                for c in self.server.connections:
                    c.notify_update(dep)

    def on_sub(self, msg):
        dr = MsgParser(msg)

        tag_name_or_id = dr.get_tag('tag')
        send_tstamp = dr.get_bool('tstamp', optional=True, default=True)
        ref = dr.get_str('ref', allow_none=True, optional=True, default='')

        has_ch_uid = 'ch' in msg
        ch_uid = msg['ch'] if has_ch_uid else None

        if not ref or ref == 'auto':
            upd_use_id = True if tag_name_or_id == '*' else isinstance(tag_name_or_id, int)
        elif ref == 'id':
            upd_use_id = True
        elif ref == 'name':
            upd_use_id = False
        else:
            raise ServerError('Parameter "ref={}" is invalid. Available: auto|name|id'.format(ref), ServerErrorCode.ProtocolError)

        # Примечание
        # Поскольку у нас отсутствует этап синхронизации, тот, кто подписывается, должен сразу получить актуальное значение.
        # Раньше он моге сделать find(), чтобы понять начальное значение, но теперь это бессмысленно, потому что
        # процессы все уже запущены, и между find/sub значение могло поменяться
        if tag_name_or_id == '*':  # subscribe all
            if self.sub_all_:
                raise ServerError('Duplicated all-subscription', ServerErrorCode.ProtocolError)

            if self.subs_map_:
                raise ServerError('All-subscription not allowed after single', ServerErrorCode.ProtocolError)

            self.sub_all_ = (send_tstamp, upd_use_id, has_ch_uid, ch_uid)

            self.log.debug('%s subscribed all', self.inst_name_)
            # см. Примечание
            for tag in self.db.all_tags_list():
                if not tag.is_void():
                    self.notify_update(tag)
        else:
            if self.sub_all_:
                raise ServerError('Cant subscribe on "{}": all-subscription is already enabled'.format(tag_name_or_id), ServerErrorCode.ProtocolError)

            tag = self.db.find_tag_by_name_or_id(tag_name_or_id)
            if not tag:
                raise ServerWarning('Cant subscribe: tag "{}" doesnt exist'.format(tag_name_or_id), 'tag_not_found')

            if tag.is_void():
                raise ServerWarning('Cant subscribe: tag "{}" is VOID'.format(tag_name_or_id), 'tag_is_void')

            if self.subs_map_ is None:
                self.subs_map_ = {}
            self.subs_map_[tag.id] = (send_tstamp, upd_use_id, has_ch_uid, ch_uid)
            self.log.debug('%s subscribed %s', self.inst_name_, tag.full_name())
            # см. Примечание
            self.notify_update(tag)

    def notify_update(self, tag):
        if self.sub_all_:
            tpl = self.sub_all_
        elif self.subs_map_:
            tpl = self.subs_map_.get(tag.id, None)
        else:
            tpl = None

        if tpl:
            send_tstamp, use_id, has_ch_uid, ch_uid = tpl
            resp = tag.vqt_dict_nonvoid(send_tstamp)
            resp['c'] = 'upd'
            resp['tag'] = tag.id if use_id else tag.full_name()
            if has_ch_uid:
                resp['ch'] = ch_uid
            self.send(resp)

    def on_find(self, msg: dict) -> None:
        dr = MsgParser(msg)

        tag_name_or_id = dr.get_tag('tag')

        tag = self.db.find_tag_by_name_or_id(tag_name_or_id)

        if tag:
            res = {'c': 'find_ok', 'id': tag.id, 'name': tag.full_name(), 'type': tag.var().type_str()}
        else:
            res = {'c': 'find_err', 'tag': tag_name_or_id}

        if 'ch' in msg:
            res['ch'] = msg['ch']

        self.send(res)

    def on_first(self, msg: dict) -> None:
        all_lst = self.db.all_tags_list()
        if all_lst:
            tag = all_lst[0]
            resp = {'c': 'next_tag', 'id': tag.id, 'name': tag.full_name(), 'type': tag.var().type_str(), 'atts': tag.attributes}
        else:
            resp = {'c': 'next_end'}

        if 'ch' in msg:
            resp['ch'] = msg['ch']

        self.send(resp)

    def on_next(self, msg: dict) -> None:
        dr = MsgParser(msg)

        tag_name_or_id = dr.get_tag('tag')

        prev = self.db.find_tag_by_name_or_id(tag_name_or_id)

        if prev:
            tag = self.db.find_tag_by_id(prev.id + 1)

            if tag:
                resp = {'c': 'next_tag', 'id': tag.id, 'name': tag.full_name(), 'type': tag.var().type_str(), 'atts': tag.attributes}
            else:
                resp = {'c': 'next_end'}
        else:
            resp = {'c': 'next_err'}

        if 'ch' in msg:
            resp['ch'] = msg['ch']

        self.send(resp)

    def send(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg

        # Ошибки dumps НЕ перехватываем. Они могут быть вызваны только проблемами в коде
        s = json.dumps(msg)     # могут быть разные ошибки. мне попадалась TypeError, но есть еще json.encoder...
        self.log.debug('S> ' + s)
        self.sendLine(s.encode('utf-8'))



#!/usr/bin/python3

from ..lib.xml_reader import XmlReader, get_node_location, load_xml_file_le, LoadException
from ..lib import variant

from . import base_dev
from . import tools
from . import mb_frame
from . import connected_values


class MbDevice(base_dev.BaseDevice):

    def __init__(self, my_xml_node, proposed_aux_name, parent_log):
        super().__init__(my_xml_node, proposed_aux_name, parent_log)

        p = XmlReader(my_xml_node)

        self.slave_nb_ = p.get_int('slaveNb', min_val=0, max_val=0xff)
        self.frames_ = []
        self.last_processed_frame_idx_ = -1

        template_name = p.get_str('template')
        template_xml = load_xml_file_le(template_name)

        self.load_frames_(template_xml.getroot())

        self.is_online_ = connected_values.ConnectedValue(self.branch('online'), variant.BoolValue(), True)

    def slave_nb(self):
        return self.slave_nb_

    def empty(self):
        return len(self.frames_) == 0

    def sync(self, services: dict):
        assert isinstance(services, dict), services
        super().sync(services)
        self.is_online_.sync(services)
        for f in self.frames_:
            f.sync(services)

    def _find_frame_by_aux_name(self, frame_aux_name):
        return tools.search(self.frames_, lambda x: x.aux_name() == frame_aux_name)

    def load_frames_(self, xml_node):
        log = self.log()

        for node in xml_node:
            frame = None
            proposed_aux_name = tools.unique_name('frame_', lambda x: not self._find_frame_by_aux_name(x))
            if node.tag == 'Registers':
                frame = mb_frame.MbFrameReg(node, proposed_aux_name, self, log)
            elif node.tag == 'Bits':
                pass
            # elif node.tag == 'Cmd':
            #    frame = MbCmdFrame( node )
            # elif node.tag == 'Opg':
            #    frame = MbFrameReadOpg( node )
            # elif node.tag == 'RegBlock'
            #    frame = MbFrameRawWriteRegs( node, this, &ok )
            else:
                raise LoadException(
                    'Unknown frame type "{}". Use "Registers, Bits, Cmd, Opg, RegBlock"'.format(node.tag), node)

            assert frame
            if frame.empty():
                log.warning('{}: Frame is empty and will be removed'.format(get_node_location(node)))
            else:
                self.frames_.append(frame)

    def on_channel_closed(self):
        self.log().debug('Channel closed and device going offline')
        self.is_online_.set(False)
        for f in self.frames_:
            f.on_channel_closed()

    def work(self, ctx):
        assert not self.empty()

        for i in range(len(self.frames_)):  # to prevent endless loop when none of frames is expired; pylint: disable=unused-variable
            self.last_processed_frame_idx_ += 1
            if self.last_processed_frame_idx_ >= len(self.frames_):
                self.last_processed_frame_idx_ = 0

            frame = self.frames_[self.last_processed_frame_idx_]
            if frame.read_write(ctx):
                return

        # log.debug ('No frames to read now')



#!/usr/bin/python3

import traceback

from ..lib.xml_reader import XmlReader

from . import time_tools
from . import tools


class BaseProtocol:
    def __init__(self, my_xml_node, comm_channel, io_ctx, proposed_name, parent_log):
        p = XmlReader(my_xml_node)

        self.comm_channel_ = comm_channel
        self.aux_name_ = p.get_str('name', proposed_name)
        self.log_ = parent_log.getChild(self.aux_name_)
        self.devices_ = []

        # comm settings
        # assert isinstance(io_ctx, BaseCtx), io_ctx
        io_ctx.silence_sec = p.get_int('silenceMs', default=0, min_val=0, max_val=10000) / 1000
        io_ctx.timeout_sec = p.get_int('timeoutMs', default=1000, min_val=50, max_val=10000) / 1000

        # run_once context:
        self.last_dev_idx_ = None
        self.is_online_ = None
        self.io_ctx_ = io_ctx
        self.offline_timer_ = time_tools.Timer()  # сколько времени уже канал закрыт
        self.offline_notified_devs_ = None  # для кого уже было вызвано .on_channel_closed(). set of aux_name (в рамках 1 канала они уникальны)
        self.open_channel_timeout_ = time_tools.Timeout(5.0, expired=True)

    def sync(self, services: dict):
        assert isinstance(services, dict), dict
        for dev in self.devices():
            dev.sync(services)

    def aux_name(self):
        return self.aux_name_

    def empty(self):
        return len(self.devices_) == 0

    def log(self):
        return self.log_

    def devices(self):
        return self.devices_

    def add_device(self, device):
        self.devices_.append(device)

    def find_device(self, aux_name):
        return tools.search(self.devices_, lambda x: x.aux_name() == aux_name)

    def run_once_safe(self, _reactor) -> None:
        log = self.log()
        channel = self.comm_channel_

        try:
            if channel.is_open():
                assert not self.empty()

                self.last_dev_idx_ = 0 if self.last_dev_idx_ is None else (self.last_dev_idx_ + 1) % len(self.devices_)

                dev = self.devices_[self.last_dev_idx_]

                dev.work(self.io_ctx_)
            else:
                # is_online_ is None    => first iteration
                # is_online_ == True    => channel crashed on last iter
                # is_online_ == False   => channel is closed and we know that already
                if self.is_online_ is None or self.is_online_:
                    just_crashed = self.is_online_
                    self.is_online_ = False

                    if just_crashed:
                        log.warning('Channel became offline')
                        self.open_channel_timeout_.start()  # на 1м цикле не надо ждать таймаут для попытки открыть канал

                    self.offline_timer_.set_now()
                    self.offline_notified_devs_ = set()

                if self.open_channel_timeout_.is_expired():
                    if channel.open(5.0, log):
                        self.is_online_ = True
                        self.last_dev_idx_ = None
                        log.info('Connection established')
                    else:
                        self.open_channel_timeout_.start()

                if not self.is_online_:  # канал мог открыться только что
                    for dev in self.devices_:
                        if dev.aux_name() not in self.offline_notified_devs_ and self.offline_timer_.secs() >= dev.channel_closed_timeout():
                            dev.on_channel_closed()
                            self.offline_notified_devs_.add(dev.aux_name())

        except Exception as e:  # exceptions in threads are not visible by default
            log.error(e)
            traceback.print_exc()

        finally:
            _reactor.callInThread(self.run_once_safe, _reactor)

    def safe_close(self) -> None:
        self.log().info('Protocol released')
        self.comm_channel_.safe_close()



import array

try:
    import fcntl
except ImportError:
    print('MOXA DI disabled for Windows')

from ..lib.xml_reader import XmlReader
from ..lib import variant

from . import time_tools
from . import connected_values


class _MoxaInput:       # pylint:disable=too-few-public-methods
    def __init__(self, node):
        p = XmlReader(node)

        self.name = p.get_str('name')
        self.index = p.get_int('index', min_val=0, max_val=3)
        self.invert = p.get_bool('invert', False)
        self.tag = connected_values.ConnectedValue(self.name, variant.BoolValue(), True)
        

class MoxaDIProtocol:
    def __init__(self, my_xml_node, proposed_name, parent_log):
        p = XmlReader(my_xml_node)

        self.aux_name_ = p.get_str('name', proposed_name)
        self.log = parent_log.getChild(self.aux_name_)
        self.inputs_ = []
        self.timeout_ = time_tools.Timeout(0.1, expired=True)

        for node in my_xml_node.findall('DI'):
            inp = _MoxaInput(node)
            self.inputs_.append(inp)

    def aux_name(self):
        return self.aux_name_

    def empty(self):
        return len(self.inputs_) == 0

    def sync(self, services: dict):
        assert isinstance(services, dict), dict
        
        for inp in self.inputs_:
            inp.tag.sync(services)

    def run_once_safe(self, _reactor) -> None:
        log = self.log

        if not self.timeout_.is_expired():
            return
    
        self.timeout_.start()

        try:
        
            with open('/dev/dio') as fd:
                for inp in self.inputs_:
                    buf = array.array('l', [inp.index, 0])
                    res = fcntl.ioctl(fd, 17, buf)
                    if res == 0:
                        input_state = 0 if buf[1] == 1 else 1
                        tag_state = input_state == 0 if inp.invert else input_state == 1
                        inp.tag.set(tag_state)
                    else:
                        log.warn('Error reading input #%s. ioctl=%s', inp.index, res)
                        inp.tag.set(False, quality=2)

        except Exception as e:  # exceptions in threads are not visible by default
            log.error(e)
            #traceback.print_exc()
            for inp in self.inputs_:
                inp.tag.set(False, quality=4)

        finally:
            _reactor.callInThread(self.run_once_safe, _reactor)

    def safe_close(self) -> None:
        self.log.info('Protocol released')



#!/usr/bin/python3


class ProtocolException(Warning):
    pass


class BadDataException(ProtocolException):
    pass


class RequestResponseMismatch(ProtocolException):
    pass


class NackException(ProtocolException):
    def __init__(self, msg, code):
        super().__init__(msg)
        self.nack_code = code



import json
import json.decoder

from twisted.protocols.basic import LineReceiver

from ..lib.stdsvcdef import IMsgOStream


#    Базовый класс для однопоточных клиентов на базе Twisted.
class VolcanoTwistedClientST(LineReceiver, IMsgOStream):
    delimiter = b'\n'

    def __init__(self, log):
        super().__init__()
        self.log = log

    def connectionMade(self):
        pass

    # Данный метод не предполагает того, что будет выброшен Exception. Причина в следующем:
    # чтобы разделить ожидаемые Exception от неожиданных (ошибок в программе), приходится
    # оговаривать, какой Exception будет считать ожидаемым, чтобы его маскировать в log.warn()
    # Но в этом случае все равно код внутри должен будет поймать свои Exception и завернуть их в ожидаемый.
    # Тогда нахрена все это надо, если можно просто сделать дисконнект?
    #
    # Вобщем, при необходимости закрыть канал следует вызвать self.close_transport_from_mt()
    # Это правильно, потому что обработчик вызывается в главном потоке.
    def on_msg_rcvd_no_exc(self, msg: dict) -> None:   # Exceptions not expected
        raise NotImplementedError()

    def rawDataReceived(self, data):
        assert False    # we never work in raw mode

    def lineReceived(self, line: bytes) -> None:
        assert isinstance(line, bytes), line

        try:
            line_s = line.decode('utf-8', 'strict').strip()
        except UnicodeDecodeError as ex:
            self.log.error('Error decoding command to utf-8: %s. Command: %s', ex, line)
            self.close_transport_from_mt()
            return

        if not line_s:
            return

        try:
            msg = json.loads(line_s)
        except json.decoder.JSONDecodeError as ex:
            self.log.error('Error processing <%s>: %s', line_s, ex)
            self.close_transport_from_mt()
            return

        try:
            self.on_msg_rcvd_no_exc(msg)
        except Exception as ex:
            self.log.error('Unexpected error while processing <%s>: %s', line_s, ex)
            raise

    # IMsgOStream
    def push_single_message(self, msg: dict):
        self.sendLine(json.dumps(msg).encode('utf-8'))

    def close_transport_from_mt(self):
        if self.transport:
            self.transport.loseConnection()



#!/usr/bin/python3


import datetime


def get_current_tstamp_utc() -> datetime.datetime:
    return datetime.datetime.utcnow()


def serialize_tstamp(tstamp_n: (datetime.datetime, None)) -> str:
    assert tstamp_n is None or isinstance(tstamp_n, datetime.datetime), tstamp_n

    return '' if tstamp_n is None else tstamp_n.strftime('%Y-%m-%dT%H:%M:%S.%f')    # %f is microsec, xxxxxx
    # return '' if tstamp_n is None else tstamp_n.isoformat(timespec='milliseconds') -- timespec appeared in 3.6


def deserialize_tstamp_w(s: (str, None)) -> (datetime.datetime, None):
    assert s is None or isinstance(s, str), s

    if not s:
        return None

    try:
        return datetime.datetime.strptime(s, '%Y-%m-%dT%H:%M:%S.%f')    # %f is microsec, xxxxxx
        # return datetime.datetime.fromisoformat(s)  -- from Python 3.7
    except ValueError as e:
        raise Warning('Invalid datetime format: {}'.format(e))



import collections
import datetime

from ..lib import variant
from ..lib.stddef import Quality
from ..lib.xml_reader import XmlReader
from ..lib.bin import word

from . import time_tools
from . import base_dev
from . import connected_values
from .merk_nzif_common import ProtocolException, NackException, BadDataException
from .nzif_defs import get_nzif_defs
from .num_ops import bcd_to_int, hi_byte, lo_byte
from .nzif_profiles import NzifProfiles, MeterInfo

NzifScaling = collections.namedtuple('NzifScaling', ('Ki', 'Ku', 'Ca', 'I_A', 'U_V', 'E_kwth', 'PQS_kwt', 'F_Hz', 'Cos'))


def _create_scaling(ki, ku, const_a, mtd) -> NzifScaling:
    i_a = 0.001 * 0.1 * float(ki) * float(mtd['ci'])  # 0.001 -перевод из мА в А
    u_v = 0.01 * float(ku)
    e_kwth = float(ki) * float(ku) / (2.0 * float(const_a))
    pqs_kwt = 0.001 * float(ku) * float(ki) * float(mtd['ks']) * 0.001  # 1ая 1/1000 - из формулы, 2ая 1/1000 - для перевода вт->кВт
    f_hz = 0.01
    cos = 0.01  # 2.4.3.23

    return NzifScaling(ki, ku, const_a, i_a, u_v, e_kwth, pqs_kwt, f_hz, cos)


def _get_uint32_ml(ba, offset):
    return (ba[offset + 0] << 24) | (ba[offset + 1] << 16) | (ba[offset + 2] << 8) | ba[offset + 3]


RWRI_SIGN_P = 'P'
RWRI_SIGN_Q = 'Q'
VAR_NAME_DATETIME = 'date_time'


class NzifDevice(base_dev.BaseDevice):

    def __init__(self, my_xml_node, proposed_aux_name, parent_log):
        super().__init__(my_xml_node, proposed_aux_name, parent_log)
        
        log = self.log()

        p = XmlReader(my_xml_node)

        self.addr_ = p.get_int('slaveNb', min_val=0, max_val=0xff)
        
        self.time_sync_delta_sec_ = p.get_int('timeSyncDeltaSec', min_val=0, max_val=600, default=30)
        self.last_sync_date_ = None     # последняя синхронизация. Объект Date()
        if self.time_sync_delta_sec_ == 0:
            log.info('Time sync disabled. Use timeSyncDeltaSec="XX" argument for Device to enable.')

        # device type
        self.type_def_ = p.get_dic('type', get_nzif_defs(log))

        # password
        pwd_str = p.get_str('pwd', '000000')
        if len(pwd_str) != 6:
            raise Exception('{}: password "{}" is invalid: should be 6 characters long'.format(p.location(), pwd_str))
        self.password_ = bytes([ord(x) for x in pwd_str])

        self.scaling_ = None

        # timeouts
        self.timeout_general_ = time_tools.Timeout(p.get_time_secs('readPeriod', 5.0, min_val=0.0, max_val=600.0), expired=True)

        # profiles
        enable_profiles = p.get_bool('profiles', default=False)
        if enable_profiles:
            mi = MeterInfo(Address=self.addr_, PulseToPhysicalUnitsFactor=1.0, MemorySize=0x10000, MaxNbOfAddrToReadAtOnce=16)
            self.profiles_ = NzifProfiles(mi)
            log.info('Profiles reading ON')
        else:
            self.profiles_ = None
            log.info('Profiles reading OFF')

        # Create values
        self.msr_1ph_ = self.type_def_['msr-1ph']
        self.msr_3ph_ = self.type_def_['msr-3ph']
        self.msr_energy_ = self.type_def_['energy']

        self.val_online_ = connected_values.ConnectedValue(self.branch('online'), variant.BoolValue(), True)

        self.values_ = {}

        for name in ('TypeCode', 'ConstA', 'Ki', 'Ku', 'Status'):
            self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.IntValue(), False)

        for x in self.msr_1ph_:
            name = x['tag']
            self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), False)

        for x in self.msr_3ph_:
            nX, n1, n2, n3 = x['tags']
            for name in (nX, n1, n2, n3):
                if name:  # nX can be None
                    self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), False)

        for x in self.msr_energy_:
            n1, n2, n3, n4 = x['tags']
            for name in (n1, n2, n3, n4):
                self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), False)

        self.values_[VAR_NAME_DATETIME] = connected_values.ConnectedValue(self.branch(VAR_NAME_DATETIME), variant.StrValue(), is_obligatory=False)


    def sync(self, services: dict):
        assert isinstance(services, dict), services
        super().sync(services)

        self.val_online_.sync(services)
        for k, v in self.values_.items():   # pylint: disable=unused-variable
            v.sync(services)

    def on_channel_closed(self):
        self.log().debug('Channel closed and device going offline')
        self.set_offline()

    def set_offline(self):
        self.val_online_.set(False)
        self.scaling_ = None
        for k, v in self.values_.items():   # pylint: disable=unused-variable
            v.invalidate(Quality.QUALITY_COMM)

    def work(self, ctx):
        log = self.log()

        if not self.timeout_general_.is_expired():
            return

        # Этот блок должен быть ДО открытия канала, потому что часть параметров (например, typeCode) считываются в критическом блоке
        # (1й try/except) и они должны быть отмечены как read уже в нем.
        for name, con_val in self.values_.items():  # pylint: disable=unused-variable
            con_val.mark_unread()

        try:
            self.open_channel(ctx)

            const_a = self.read_type_info(ctx)

            if self.scaling_ is None:
                ki, ku = self.read_ki_ku(ctx)
                self.scaling_ = _create_scaling(ki, ku, const_a, self.type_def_)

            self.read_status(ctx)

        except (ConnectionError, TimeoutError, ProtocolException) as e:
            log.warning(e)
            self.set_offline()
            self.timeout_general_.start()
            return

        self.val_online_.set(True)

        try:  # finally -> restart timeout
            for x in self.msr_3ph_:
                nX, n1, n2, n3 = x['tags']
                f = x['f']
                p1, p2, p3 = x['rwri']
                comment = 'Read {} {} {}'.format(n1, n2, n3)
                try:
                    self.read_3phase(ctx, getattr(self.scaling_, f), p1, p2, p3, comment, (nX, n1, n2, n3))
                except (ConnectionError, TimeoutError, ProtocolException) as e:
                    log.warning('Error reading {}: {}'.format(comment, e))
                    if not ctx.is_open():
                        raise

            for x in self.msr_1ph_:
                n = x['tag']
                f = x['f']
                p1, p2, p3 = x['rwri']

                comment = 'Read {}'.format(n)
                try:
                    self.read_1phase(ctx, getattr(self.scaling_, f), p1, p2, p3, comment, n)
                except (ConnectionError, TimeoutError, ProtocolException) as e:
                    log.warning('Error reading {}: {}'.format(comment, e))
                    if not ctx.is_open():
                        raise

            for x in self.msr_energy_:
                n1, n2, n3, n4 = x['tags']
                arr = x['arr']
                comment = 'Read energy {} {} {} {}'.format(n1, n2, n3, n4)
                try:
                    self.read_energy(ctx, arr, comment, (n1, n2, n3, n4))
                except (ConnectionError, TimeoutError, ProtocolException) as e:
                    log.warning('Error reading {}: {}'.format(comment, e))
                    if not ctx.is_open():
                        raise
            # timesync
            try:
                self.time_sync(ctx)
            except (ConnectionError, TimeoutError, ProtocolException) as ex:
                log.warning('Error during time sync: {}'.format(ex))
                if not ctx.is_open():
                    raise

            # profiles
            try:
                if self.profiles_:
                    self.profiles_.read_profiles(ctx, 2, log)
            except (ConnectionError, TimeoutError, ProtocolException) as ex:
                log.warning('Error during reading profiles: {}'.format(ex))
                if not ctx.is_open():
                    raise

        except (ConnectionError, TimeoutError, ProtocolException) as e:
            # Here we get errors only from Ph1, Ph3, En reads when ctx is not recoverable.
            # Isonline remains True
            # log.warn already called
            for name, con_val in self.values_.items():
                if not con_val.is_read():
                    con_val.invalidate(Quality.QUALITY_COMM)
        finally:
            self.timeout_general_.start()

    def open_channel(self, ctx):
        log = self.log_

        req = bytes([self.addr_, 0x01]) + self.password_

        try:
            ctx.send_rcv(req, None, 'Open channel', self.log())
        except NackException as e:
            if e.nack_code == 3:
                log.warning('Nack 3 in open channel often caused by invalid password. Current password: {}'.format(self.password_))
            raise

    # rval: ConstA / raise
    def read_type_info(self, ctx):
        log = self.log()
        # 2.4.3.24
        # Исполнение, константа А
        # Команда предназначена для чтения варианта исполнения счетчика, установленного на
        # заводе-изготовителе. Команда введена в счетчики СЭТ-4ТМ.02 начиная с V14.ХХ.ХХ.
        req = bytes([self.addr_, 0x08, 0x12])

        ctx.send_rcv(req, 3, 'Read type info', self.log())  # * is ok

        const_a_code = ctx.data(1) & 0x0f
        const_a_map = {
            0: 5000,
            1: 25000,
            2: 1250,
            3: 6250,
            4: 500,
            5: 250,
            6: 6400
        }
        const_a = const_a_map.get(const_a_code, None)
        if const_a is None:
            raise BadDataException('{}: Unknown const A code: {}'.format('Read type info', const_a_code))

        type_code = (ctx.data(2) & 0xf0) >> 4
        if self.type_def_['type_code'] != type_code:
            raise BadDataException(
                'Type code mismatch. Expected {}, got {}'.format(self.type_def_['type_code'], type_code))

        log.debug("TypeCode=%s, ConstA=%s", type_code, const_a)

        self.values_['TypeCode'].set(type_code)
        self.values_['ConstA'].set(const_a)

        return const_a

    # rval: ki,ku / raise
    def read_ki_ku(self, ctx):
        # Ki Ku  - 2.4.3.3
        log = self.log()

        req = bytes((self.addr_, 0x08, 0x02))

        ctx.send_rcv(req, 10, 'Read KiKu', self.log())

        ku = word(ctx.data(0), ctx.data(1))
        ki = word(ctx.data(2), ctx.data(3))

        log.debug('Ki=%s, Ku=%s', ki, ku)

        self.values_['Ki'].set(ki)
        self.values_['Ku'].set(ku)

        return ki, ku

    def read_status(self, ctx) -> None:
        # 2.4.3.14
        log = self.log()

        req = bytes((self.addr_, 0x08, 0x0a))
        nb_bytes_expected = self.type_def_['st_bytes']

        ctx.send_rcv(req, nb_bytes_expected, 'Read status', self.log())

        d = ctx.data()
        status = 0
        for i in range(nb_bytes_expected):
            status += d[i] << (8 * i)

        log.debug('Status={:02x}'.format(status))

        self.values_['Status'].set(status)

    # при групповом чтении измерений возвращает элемент
    # item:
    #	0 = сумма фаз
    #	1-3 = по указанной фазе
    # rval: (value, sign_p, sign_q)
    @staticmethod
    def get_group_read_item(data, item):
        # 2.4.3.23
        # В ответ на запрос счетчик возвращает слово из трех байт в двоичном коде. Два старших
        # бита старшего байта указывают положение вектора полной мощности и должны маскироваться
        # для получения модуля значения параметра. В счетчиках СЭТ-4ТМ.02,03М, ПСЧ-3,4ТМ.05М
        # при чтении мощности потерь старшие биты указывают положение вектора полной мощности потерь.
        hi_b = data[item * 3 + 0]
        hi_b_data = hi_b & 0x3f
        hi_b_p_sgn = hi_b & 0x80
        hi_b_q_sgn = hi_b & 0x40

        value = (hi_b_data << 16) | (data[item * 3 + 1] << 8) | data[item * 3 + 2]
        sign_p = 1 if hi_b_p_sgn == 0 else -1
        sign_q = 1 if hi_b_q_sgn == 0 else -1

        return value, sign_p, sign_q

    def time_sync(self, ctx):
        # 2.4.1.1.1
        log = self.log_
        comment = 'Read time'
        con_val = self.values_[VAR_NAME_DATETIME]

        req = bytes([self.addr_, 0x04, 0x00])

        try:
            ctx.send_rcv(req, 8, comment, log)
        except NackException:
            con_val.invalidate(Quality.QUALITY_NA)
            raise
            
        # Замер системного времени надо сделать максимально близко к чтению времени из счетчика
        sys_time = datetime.datetime.now()

        try:
            ss = bcd_to_int(ctx.data(0))        # !ValueError
            mm = bcd_to_int(ctx.data(1))
            hh = bcd_to_int(ctx.data(2))
            D = bcd_to_int(ctx.data(4))
            M = bcd_to_int(ctx.data(5))
            Y = bcd_to_int(ctx.data(6)) + 2000
            
            dtm = datetime.datetime(Y, M, D, hh, mm, ss)        # !ValueError
        except ValueError as ex:
            log.warn(ex)
            con_val.invalidate(Quality.QUALITY_NA)
            raise ProtocolException(str(ex))
        
        con_val.set(str(dtm))

        # compare 
        delta = sys_time - dtm
        delta_secs = int(delta.total_seconds())

        log.debug('Datetime: Meter=%s, System=%s, Delta=%s sec', dtm, sys_time, delta_secs)

        if self.time_sync_delta_sec_ == 0 or self.time_sync_delta_sec_ > abs(delta_secs):
            log.debug('Time sync disabled for device')
            return
        
        log.debug('Time sync needed: Delta is %s sec [Max allowed = %s sec]', abs(delta_secs), self.time_sync_delta_sec_)
        
        if self.last_sync_date_ is not None and self.last_sync_date_ == datetime.date.today():
            log.debug('Time sync already done today')
            return
            
        if delta_secs > 120:
            secs_to_add = 120
        elif delta_secs < -120: 
            secs_to_add = -120
        else:
            secs_to_add = delta_secs
            
            
        # Ограничением выполнения команды является требование коррекции времени без перехода
        # в следующий или предыдущий час. Если, тем не менее, запрашивается коррекция времени
        # с переходом в следующий или предыдущий час, то счетчик возвращает в байте состояния обмена
        # код 01h (недопустимая команда или параметр).
        # Поскольку коррекция идет макс на 120 сек (2 мин), то проверим минуты
        new_time = dtm + datetime.timedelta(seconds=secs_to_add)
        if dtm.hour != new_time.hour:
            log.debug('Time sync now would result in hour change. Postponing')
            return
            
        # Тут могут быть отрицательные значения, но для нас это неважно: int32 для отрицательных
        # значений будет FF FF FF XY, а i16 FF XY, то есть отбрасывание старших
        # байтов валидно.
        req = bytes([self.addr_, 0x03, 0x0b, hi_byte(secs_to_add), lo_byte(secs_to_add)])
        comment = 'Adjust time'
            
        try:
            ctx.send_rcv(req, None, comment, log)
        except NackException as ex:
            # con_val.invalidate(Quality.QUALITY_NA) Это была запись, нет нужды инвалидировать текущее время
            # Часы уже корректировались, запомним это. Такой вариант бывает, если Ридер запустить повторно, ведь я пока не сохраняю в БД даты последней коррекции
            if ex.nack_code == 0x04:
                log.warn('Time sync failed with nack code 4 (already synced today)')
                return
            # Как было описано выше, этот код означает, что была попытка перевести часы в другой час. Хотя у меня и есть
            # защита от этого, но кто знает.
            elif ex.nack_code == 0x01:
                log.warn('Time sync failed with nack code 1 (attempt to change an hour). Just repeat later.')
                return
            else:
                raise
            
        log.info('Time successfully synced')
        self.last_sync_date_ = datetime.date.today()
        

    @staticmethod
    def msr_code_rwri(mode, param, phase):
        return (mode << 4) | (param << 2) | phase

    def read_3phase(self, ctx, scaling_f, rwri_p1, rwri_p2, rwri_sign, comment, targets_tuple):
        # 2.4.3.28
        log = self.log_

        req = bytes([self.addr_, 0x08, 0x16, NzifDevice.msr_code_rwri(rwri_p1, rwri_p2, 0)])

        try:
            ctx.send_rcv(req, 12, comment, log)
        except NackException as e:
            q = Quality.QUALITY_NA_TEMP if e.nack_code == 0x07 else Quality.QUALITY_NA
            for n in targets_tuple:
                if n:
                    self.values_[n].invalidate(q)
            raise

        sum_raw = 0.0
        if targets_tuple[0] is not None:
            sum_raw, sign_p, sign_q = NzifDevice.get_group_read_item(ctx.data(), 0)
            if rwri_sign == RWRI_SIGN_P: sum_raw = sum_raw * sign_p
            if rwri_sign == RWRI_SIGN_Q: sum_raw = sum_raw * sign_q

        ph1_raw, sign_p, sign_q = NzifDevice.get_group_read_item(ctx.data(), 1)
        if rwri_sign == RWRI_SIGN_P: ph1_raw = ph1_raw * sign_p
        if rwri_sign == RWRI_SIGN_Q: ph1_raw = ph1_raw * sign_q

        ph2_raw, sign_p, sign_q = NzifDevice.get_group_read_item(ctx.data(), 2)
        if rwri_sign == RWRI_SIGN_P: ph2_raw = ph2_raw * sign_p
        if rwri_sign == RWRI_SIGN_Q: ph2_raw = ph2_raw * sign_q

        ph3_raw, sign_p, sign_q = NzifDevice.get_group_read_item(ctx.data(), 3)
        if rwri_sign == RWRI_SIGN_P: ph3_raw = ph3_raw * sign_p
        if rwri_sign == RWRI_SIGN_Q: ph3_raw = ph3_raw * sign_q

        sum_physical = sum_raw * scaling_f
        ph1_physical = ph1_raw * scaling_f
        ph2_physical = ph2_raw * scaling_f
        ph3_physical = ph3_raw * scaling_f

        log.debug('%s: sum_raw=%s, ph1_raw=%s, ph2_raw=%s, ph3_raw=%s, sum=%s, ph1=%s, ph2=%s, ph3=%s',
                  comment, sum_raw, ph1_raw, ph2_raw, ph3_raw, sum_physical, ph1_physical, ph2_physical, ph3_physical)

        if targets_tuple[0] is not None:
            self.values_[targets_tuple[0]].set(sum_physical)
        self.values_[targets_tuple[1]].set(ph1_physical)
        self.values_[targets_tuple[2]].set(ph2_physical)
        self.values_[targets_tuple[3]].set(ph3_physical)

    def read_1phase(self, ctx, scaling_f, rwri_p1, rwri_p2, rwri_p3, comment, target):
        # 2.4.3.28
        log = self.log_

        req = bytes((self.addr_, 0x08, 0x11, NzifDevice.msr_code_rwri(rwri_p1, rwri_p2, rwri_p3)))

        try:
            ctx.send_rcv(req, 3, comment, log)
        except NackException as e:
            q = Quality.QUALITY_NA_TEMP if e.nack_code == 0x07 else Quality.QUALITY_NA
            self.values_[target].invalidate(q)
            raise

        val_raw, sign_p, sign_q = NzifDevice.get_group_read_item(ctx.data(), 0) # pylint: disable=unused-variable
        val_phys = val_raw * scaling_f

        log.debug('%s: raw=%s, phys=%s', comment, val_raw, val_phys)

        self.values_[target].set(val_phys)

    def read_energy(self, ctx, array_nb, comment, tags_tuple):
        log = self.log_

        assert self.scaling_ is not None

        month_nb = 0
        tariff_nb = 0

        req = bytes([self.addr_, 0x05, (array_nb << 4) | month_nb, tariff_nb])

        try:
            ctx.send_rcv(req, 16, comment, log)
        except NackException as e:
            q = Quality.QUALITY_NA_TEMP if e.nack_code == 0x07 else Quality.QUALITY_NA
            for n in tags_tuple:
                self.values_[n].invalidate(q)
            raise

        f = self.scaling_.E_kwth
        eap = _get_uint32_ml(ctx.data(), 0) * f
        ean = _get_uint32_ml(ctx.data(), 4) * f
        erp = _get_uint32_ml(ctx.data(), 8) * f
        ern = _get_uint32_ml(ctx.data(), 12) * f

        log.debug("Eap=%s, Ean=%s, Erp=%s, Ern=%s", eap, ean, erp, ern)

        self.values_[tags_tuple[0]].set(eap)
        self.values_[tags_tuple[1]].set(ean)
        self.values_[tags_tuple[2]].set(erp)
        self.values_[tags_tuple[3]].set(ern)




from .stddef import ValueType


class Variant:
    vt_ = None
    value_ = None

    def vt(self) -> ValueType:  # pylint: disable=invalid-name
        return self.vt_

    def type_str(self) -> str:
        return self.vt_.stringify()

    def get_value(self):
        return self.value_

    # try set from given raw value. throw warning if conversion is required.
    def set_value_w(self, value) -> None:
        raise NotImplementedError()

    # None is not accepted
    def parse_value_w(self, s: str):    # pylint: disable=invalid-name
        raise NotImplementedError()

    def clone(self) -> 'Variant':
        raise NotImplementedError()

    # True if both not none & both not VOID & both have equal types
    def __eq__(self, x):
        if self.vt_ == x.vt_:
            v1 = self.get_value()
            v2 = x.get_value()

            return v1 is not None and v2 is not None and v1 == v2
        else:
            return False

    def __ne__(self, x):
        return not self.__eq__(x)

    def same_type(self, other: 'Variant'):
        assert isinstance(other, Variant), other

        return self.vt_ == other.vt_

    def __str__(self):
        return str(self.value_)

    def val_equal(self, value, none_eq_none: bool):
        my_val = self.value_

        if my_val is None and value is None:
            return none_eq_none

        return my_val == value  # None is not equal to 0/'' so no need to check if one argument is None

    def is_void(self) -> bool:
        return self.vt_ == ValueType.VT_VOID

    # asserts if types differ
    def copy_from_same_type(self, src: 'Variant'):
        assert isinstance(src, Variant), src
        assert self.vt_ == src.vt_, (self, src)

        self.value_ = src.value_

    @staticmethod
    def from_vt(vt: ValueType) -> 'Variant':
        if vt == ValueType.VT_VOID:
            return VOID_VALUE
        elif vt == ValueType.VT_BOOL:
            return BoolValue()
        elif vt == ValueType.VT_INT:
            return IntValue()
        elif vt == ValueType.VT_FLOAT:
            return FloatValue()
        else:
            assert vt == ValueType.VT_STR, vt
            return StrValue()

    # none is not supported
    @staticmethod
    def try_from_raw_value(value) -> ('Variant', None):
        if isinstance(value, bool):
            return BoolValue(value)
        if isinstance(value, int):
            return IntValue(value)
        if isinstance(value, float):
            return FloatValue(value)
        if isinstance(value, str):
            return StrValue(value)
        return None


class VoidValue(Variant):
    vt_ = ValueType.VT_VOID

    def clone(self):
        return self  # void value is immutable

    def set_value_w(self, value):
        raise Warning('VOID type cant set value: {}'.format(value))

    def parse_value_w(self, s: str):
        assert isinstance(s, str), s
        raise Warning('VOID type cant set value: {}'.format(s))


VOID_VALUE = VoidValue()


class BoolValue(Variant):
    vt_ = ValueType.VT_BOOL

    def __init__(self, value=None):
        if value is not None:
            self.set_value_w(value)

    def clone(self):
        rs = BoolValue()
        rs.value_ = self.value_  # this approach is faster than XXValue(self.value) coz we dont need to perform type checking
        return rs

    def set_value_w(self, value):
        if isinstance(value, bool):
            self.value_ = value
        elif value is None:
            self.value_ = None
        elif isinstance(value, int):
            if value == 1:
                self.value_ = True
            elif value == 0:
                self.value_ = False
            else:
                raise Warning('Value {} cannot be interpreted as bool'.format(value))
        else:
            raise Warning('Value {} cannot be interpreted as bool'.format(value))

    # none is not accepted
    def parse_value_w(self, s: str):
        assert isinstance(s, str), s
        if s == '1':
            self.value_ = True
        elif s == '0':
            self.value_ = False
        else:
            raise Warning('String "{}" cannot be parsed to bool. Use 1 and 0 values'.format(s))


class IntValue(Variant):
    vt_ = ValueType.VT_INT

    def __init__(self, value=None):
        if value is not None:
            self.set_value_w(value)

    def clone(self):
        rs = IntValue()
        rs.value_ = self.value_  # this approach is faster than XXValue(self.value) coz we dont need to perform type checking
        return rs

    def set_value_w(self, value):
        # in Python bool is subclass of int!
        if isinstance(value, bool):
            self.value_ = 1 if value else 0
        elif value is None or isinstance(value, int):
            self.value_ = value
        elif isinstance(value, float):
            if not value.is_integer():
                raise Warning('Value {} cannot be interpreted as int'.format(value))

            self.value_ = int(value)
        else:
            raise Warning('Value {} cannot be interpreted as int'.format(value))

    # none is not accepted
    def parse_value_w(self, s: str):
        assert isinstance(s, str), s
        try:
            self.value_ = int(s)
        except ValueError:
            raise Warning('Value {} cannot be interpreted as int'.format(s))


class FloatValue(Variant):
    vt_ = ValueType.VT_FLOAT

    def __init__(self, value=None):
        if value is not None:
            self.set_value_w(value)

    def clone(self):
        rs = FloatValue()
        rs.value_ = self.value_  # this approach is faster than XXValue(self.value) coz we dont need to perform type checking
        return rs

    def set_value_w(self, value):
        if value is None or isinstance(value, float):
            self.value_ = value
        elif isinstance(value, int):
            # here we also get bool!
            self.value_ = float(value)
        else:
            raise Warning('Value {} cannot be interpreted as float'.format(value))

    # none is not accepted
    def parse_value_w(self, s: str):
        assert isinstance(s, str), s
        try:
            self.value_ = float(s)
        except ValueError:
            raise Warning('Value {} cannot be interpreted as float'.format(s))


class StrValue(Variant):
    vt_ = ValueType.VT_STR

    def __init__(self, value=None):
        if value is not None:
            self.set_value_w(value)

    def clone(self):
        rs = StrValue()
        rs.value_ = self.value_  # this approach is faster than XXValue(self.value) coz we dont need to perform type checking
        return rs

    def set_value_w(self, value):
        if value is None or isinstance(value, str):
            self.value_ = value
        else:
            raise Warning('Value {} cannot be interpreted as str'.format(value))

    # none is not accepted
    def parse_value_w(self, s: str):
        assert isinstance(s, str), s
        self.value_ = s



import enum
from . import stdint


'''
    раньше было 270, но я лично видел ответы на 100 функцию длиной 275
    байт (rtu + crc)
    Данное значение не дб точным, оно дб НЕ МЕНЬШЕ реального.
    Оно используется для выделения буферов.
    At this moment longest is request to write words:
    Slave(1),Fn(1),Addr(2),NbWords(2),NbBytes(1),DATA,Crc(2)
    Data can be up to 254bytes - cause NbBytes(1) field
    Total is 254+9 = 263 bytes
    Данная длина покрывает максимальный запрос + CRC + TCP Header, то есть по
    большому счету - несуществующий запрос
    (и crc, и tcp). Но в некоторых программах это испольуется (буфер один на rtu
    и tcp)
'''     # pylint: disable=pointless-string-statement

MAX_FRAME_SIZE_BYTES = 306
MAX_READ_WORDS_NB = 127
MAX_WRITE_WORDS_NB = 127
MAX_READ_BITS_NB = 2000  # See JBus_Modbus.pdf page 40
MAX_WRITE_BITS_NB = 1968  # See JBus_Modbus.pdf page 44

MB_FN_READ_OUT_BITS_1 = 1
MB_FN_READ_IN_BITS_2 = 2
MB_FN_READ_OUT_WORDS_3 = 3
MB_FN_READ_IN_WORDS_4 = 4
MB_FN_WRITE_BIT_5 = 5
MB_FN_WRITE_WORD_6 = 6
MB_FN_WRITE_BITS_15 = 15
MB_FN_WRITE_WORDS_16 = 16

MB_EXC_ILLEGAL_FN_CODE = 0x01
MB_EXC_ILLEGAL_DATA_ADDRESS = 0x02
# Величина, содержащаяся в поле данных запроса, является недопустимой величиной для подчиненного.
MB_EXC_ILLEGAL_DATA_VALUE = 0x03
MB_EXC_SERVER_FAILURE = 0x04


class MbLimits(enum.IntEnum):
    MAX_FRAME_SIZE_BYTES = 306
    MAX_READ_WORDS_NB = 127
    MAX_WRITE_WORDS_NB = 127
    MAX_READ_BITS_NB = 2000  # See JBus_Modbus.pdf page 40
    MAX_WRITE_BITS_NB = 1968  # See JBus_Modbus.pdf page 44


class MbFunctionCode(enum.IntEnum):
    MB_FN_READ_OUT_BITS_1 = 1
    MB_FN_READ_IN_BITS_2 = 2
    MB_FN_READ_OUT_WORDS_3 = 3
    MB_FN_READ_IN_WORDS_4 = 4
    MB_FN_WRITE_BIT_5 = 5
    MB_FN_WRITE_WORD_6 = 6
    MB_FN_WRITE_BITS_15 = 15
    MB_FN_WRITE_WORDS_16 = 16


'''
    Подчиненный принял запрос и обрабатывает его, но это требует много
    времени. Этот ответ предохраняет главного от генерации ошибки тайм-аута.
'''
MB_EXC_ACKNOWLEDGE = 0x05
'''
    Подчиненный занят обработкой команды. Главный должен повторить
    позже, когда подчиненный освободится.
'''
MB_EXC_SERVER_BUSY = 0x06
'''
    Подчиненный не может выполнить программную функцию, принятую в
    запросе. Этот код возвращается для неудачного программного запроса,
    использующего функции с номерами 13 или 14. Главный должен запросить
    диагностическую информацию или информацию об ошибках с подчиненного.
'''
MB_EXC_CANT_PROCESS_FN = 0x07
'''
    Подчиненный пытается читать расширенную память, но обнаружил ошибку
    паритета. Главный может повторить запрос, но обычно в таких случаях требуется ремонт.
'''
MB_EXC_MEM_PARITY = 0x08
MB_EXC_GATEWAY_PATHS_NA = 0x0A
MB_EXC_TARGETED_DEVICE_FAILED_TO_RESPOND = 0x0B


class MbExceptionCode(enum.IntEnum):
    MB_EXC_ACKNOWLEDGE = 0x05
    MB_EXC_SERVER_BUSY = 0x06
    MB_EXC_CANT_PROCESS_FN = 0x07
    MB_EXC_MEM_PARITY = 0x08
    MB_EXC_GATEWAY_PATHS_NA = 0x0A
    MB_EXC_TARGETED_DEVICE_FAILED_TO_RESPOND = 0x0B


class MbException(Exception):
    def __init__(self, code, msg=None):
        if msg:
            super().__init__('Modbus exception 0x{:x}: {}'.format(code, msg))
        else:
            super().__init__('Modbus exception 0x{:x}'.format(code))
        self.code = code


class ErrorCode(enum.IntEnum):
    BadData = 1
    UnknownFn = 2
    UnknownExc = 3
    BadCrc = 4
    NotEnoughSpace = 5  # при формировании сообщения буфер оказался недостаточным
    MismatchTrId = 6
    MismatchSlave = 7
    MismatchFn = 8
    MismatchData = 9


class MbError(Exception):
    def __init__(self, code, msg=None):
        # str(code) gives variable name
        if msg:
            super().__init__('Modbus error <{} {}>. {}'.format(code, str(code), msg))
        else:
            super().__init__('Modbus error <{} {}>'.format(code, str(code)))
        self.code = code


# (fmt_id, fmt_size_words, fmt_min_n|None, fmt_max_n|None, fmt_struct_format, val_type)
MB_FORMATS_LIST = (
    ('i16',      1, stdint.MIN_I16, stdint.MAX_I16,     '>h', int),     # noqa: E241
    ('ui16',     1, 0,              stdint.MAX_UI16,    '>H', int),     # noqa: E241
    ('i32_be',   2, stdint.MIN_I32, stdint.MAX_I32,     '>i', int),     # noqa: E241
    ('ui32_be',  2, 0,              stdint.MAX_UI32,    '>I', int),     # noqa: E241
    ('i64_be',   4, stdint.MIN_I64, stdint.MAX_I64,     '>q', int),     # noqa: E241
    ('ui64_be',  4, 0,              stdint.MAX_UI64,    '>Q', int),     # noqa: E241
    ('f32_be',   2, None,           None,               '>f', float),   # noqa: E241
)


MB_FORMATS_DICT = {}
for f in MB_FORMATS_LIST:
    MB_FORMATS_DICT[f[0]] = f



import re

from ..lib.stddef import ValueType

RE_BIT = re.compile(r'(.+) bit (\d+)')


class Expr:

    @staticmethod
    def load_w(expr_str: str, target_tag, db):  # pylint: disable=invalid-name
        assert isinstance(expr_str, str), expr_str

        m = RE_BIT.fullmatch(expr_str)
        if m:
            return ExprBit(expr_str, target_tag, db, m)
        raise Warning('Expression "{}" is invalid. Possible options: "XXX bit N"'.format(expr_str))

    # list of tags
    def sources(self) -> tuple:
        raise NotImplementedError()

    # rval: (value, quality)
    def exec_safe(self, log) -> tuple:
        raise NotImplementedError()


class ExprBit(Expr):
    def __init__(self, expr_str: str, target_tag, db, m):   # pylint: disable=invalid-name, unused-argument
        vt = target_tag.var().vt()
        if vt != ValueType.VT_BOOL:
            raise Warning('Target tag should be BOOL for bit expression. Tag {} is {}'.format(target_tag.full_name(), vt))

        source_name = m.group(1)
        bit_idx_str = m.group(2)

        source = db.resolve_path_w(target_tag, source_name)
        if source.var().vt() != ValueType.VT_INT:
            raise Warning('Source tag should be INT for bit expression. Tag {} is {}'.format(source.full_name(), source.var().vt()))

        try:
            bit_idx = int(bit_idx_str)
        except ValueError as e:
            raise Warning('{} is not a valid bit index: {}'.format(m.group(1), e))

        if bit_idx < 0 or bit_idx > 63:
            raise Warning('{} is not a valid bit index'.format(bit_idx))

        self.bit_mask_ = 1 << bit_idx
        self.src_tag_ = source

    def sources(self) -> tuple:
        return (self.src_tag_,)

    def exec_safe(self, log) -> tuple:
        src_v = self.src_tag_.var().get_value()
        src_q = self.src_tag_.quality

        if src_v is None:
            return None, src_q

        assert isinstance(src_v, int), src_v
        return bool(src_v & self.bit_mask_), src_q



BUILD_ON = False
tags = []


def enable_build() -> None:
    global BUILD_ON     # pylint: disable=global-statement
    BUILD_ON = True


def build_tag(branch: (str, None), long_name: str, vt: str) -> None:
    assert branch is None or isinstance(branch, str), branch
    assert isinstance(long_name, str), long_name
    assert isinstance(vt, str), vt

    if BUILD_ON:
        if branch:
            full_name = branch + '.' + long_name
        else:
            full_name = long_name

        tags.append((full_name, vt))


def flush_build() -> None:
    assert BUILD_ON
    """
    print('<?xml version="1.0" encoding="UTF-8"?>')
    print('<doc>')
    for name, vt in tags:
        print('  <Tag name="{}" type="{}" />'.format(name, vt))
    print('</doc>')
    """
    for name, vt in tags:
        print('tag {} {}'.format(name, vt))



import json
import json.decoder
import threading

from twisted.internet import task
from twisted.protocols.basic import LineReceiver

from ..lib.stdsvcdef import IMsgOStream


#    Базовый класс для многопоточных клиентов на базе Twisted.
#    Реализует внутренние очереди сообщений
class VolcanoTwistedClientMT(LineReceiver, IMsgOStream):
    delimiter = b'\n'

    def __init__(self, log):
        super().__init__()
        self.log = log
        self.outbox_ = []  # list of msgs
        self.lock_ = threading.Lock()
        self.looping_call_handler_ = None

    def on_msg_rcvd_mt(self, msg: dict) -> None:
        raise NotImplementedError()

    def connectionMade(self):
        self.looping_call_handler_ = task.LoopingCall(self.looping_call_)  # looping_call_.stop() will stop the looping calls
        self.looping_call_handler_.start(0.01)

    def looping_call_(self):
        self.lock_.acquire()
        try:
            for msg in self.outbox_:
                self.sendLine(json.dumps(msg).encode('utf-8'))
            self.outbox_ = []
        finally:
            self.lock_.release()

    def lineReceived(self, line: bytes) -> None:
        assert isinstance(line, bytes), line

        try:
            line_s = line.decode('utf-8', 'strict').strip()
        except UnicodeDecodeError as ex:
            self.log.error('Error decoding command to utf-8: %s. Command: %s', ex, line)
            self.close_transport_from_mt()
            return

        if not line_s:
            return

        try:
            msg = json.loads(line_s)
        except json.decoder.JSONDecodeError as ex:
            self.log.error('Error processing <%s>: %s', line_s, ex)
            self.close_transport_from_mt()
            return

        try:
            self.on_msg_rcvd_mt(msg)
        except Exception as ex:
            self.log.error('Unexpected error while processing <%s>: %s', line_s, ex)
            raise

    # IMsgOStream
    def push_single_message(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg
        self.lock_.acquire()
        try:
            self.outbox_.append(msg)
        finally:
            self.lock_.release()

    def close_transport_from_mt(self):
        if self.transport:
            self.transport.loseConnection()



# Здесь собраны "классические" определения сервисов


# Все нижеописанные сервисы подразумевают наличие канала отправки сообщений, причем неважно, на чем реализованного
class IMsgOStream:
    def push_single_message(self, msg: dict) -> None:
        raise NotImplementedError()

    def push_multiple_messages(self, messages: (list, tuple)) -> None:
        assert isinstance(messages, (list, tuple)), messages
        for m in messages:
            self.push_single_message(m)

    def flush(self) -> None:
        pass


class IHandshakeService:
    def salute(self, inst_name: str) -> None:
        raise NotImplementedError()


class ISetValueService:
    def set_tag(self, tag_name_or_id: (str, int), value, quality: int = 0, tstamp: ('datetime.datetime', None) = None) -> None:
        raise NotImplementedError()


class IFindTagHandler:
    def on_find_tag_ok(self, tag_id: int, tag_name: str, vt: str, user_data):
        raise NotImplementedError()

    def on_find_tag_err(self, tag_name_or_id: (int, str), user_data):
        raise NotImplementedError()


class IFindTagService:
    def find_tag(self, tag_name_or_id: (str, int), handler: IFindTagHandler, user_data=None) -> None:
        raise NotImplementedError()


class IEchoService:
    def echo(self, callback, *args, **kwargs) -> None:
        raise NotImplementedError()


class ITagUpdateHandler:
    def on_tag_updated(self, tag_name_or_id: (int, str), val_raw, quality: int, tstamp_n: ('datetime.datetime', None)):
        raise NotImplementedError()


class ISubscribeService:
    def subscribe(self, tag_name_or_id: (int, str), send_tstamp: bool, handler: ITagUpdateHandler) -> None:
        raise NotImplementedError()


class ISubscribeAllService:
    def subscribe_all(self, send_tstamp: bool, use_tag_id: bool, handler: ITagUpdateHandler):
        raise NotImplementedError()


class ITagIterHandler:
    def on_next_tag(self, tag_id: int, tag_name: str, vt: str, atts: dict):
        raise NotImplementedError()

    def on_next_end(self):
        raise NotImplementedError()

    def on_next_err(self):
        raise NotImplementedError()


class ITagIterService:
    def first(self, handler: ITagIterHandler):
        raise NotImplementedError()

    def next(self, tag_name_or_id: (int, str), handler: ITagIterHandler):
        raise NotImplementedError()



#!/usr/bin/python3
import enum


class ServerErrorCode(enum.Enum):
    BadVersion = 'bad_version'
    NoHandshake = 'no_handshake'
    NoSlot = 'no_slot'
    SyntaxError = 'bad_syntax'
    ProtocolError = 'protocol_error'


class ServerError(UserWarning):
    def __init__(self, msg: str, code: ServerErrorCode):
        assert isinstance(msg, str), msg
        assert isinstance(code, ServerErrorCode), code

        super().__init__(msg)
        self.code = code


class ServerWarning(UserWarning):
    def __init__(self, msg: str, code: str):
        assert isinstance(msg, str), msg
        assert isinstance(code, str), code

        super().__init__(msg)
        self.code = code



from ..lib.bin import get_crc16_as_bytes
from ..lib.modbus import MbError, ErrorCode, MbFunctionCode, MbLimits, MbException


def word_ml(ba: (bytes, bytearray), offset: int) -> int:
    assert isinstance(ba, (bytes, bytearray)), ba
    assert isinstance(offset, int), offset
    assert len(ba) > (offset + 1)

    return (ba[offset] << 8) | ba[offset + 1]


def lo(w: int) -> int:
    assert isinstance(w, int), w
    return w & 0xff


def hi(w: int) -> int:
    assert isinstance(w, int), w
    return (w >> 8) & 0xff


def _wrap_req(trans_id: (int, None), rtu_data: (bytes, bytearray)) -> bytes:
    assert trans_id is None or isinstance(trans_id, int), trans_id
    assert isinstance(rtu_data, (bytes, bytearray)), rtu_data

    if trans_id is None:  # rtu
        return bytes(rtu_data) + get_crc16_as_bytes(rtu_data)
    else:
        if trans_id < 0 or trans_id > 0xffff:
            raise MbError(ErrorCode.BadData)

        return bytes([hi(trans_id), lo(trans_id), 0, 0, hi(len(rtu_data)), lo(len(rtu_data))]) + bytes(rtu_data)


# Raise MbError if parameters are incorrect
def build_req_read_words(trans_id_none: int, slave_nb: int, fn_nb: int, addr: int, nb_words: int) -> bytes:
    assert trans_id_none is None or isinstance(trans_id_none, int), trans_id_none
    assert isinstance(slave_nb, int), slave_nb
    assert isinstance(fn_nb, int), fn_nb
    assert isinstance(addr, int), addr
    assert isinstance(nb_words, int), nb_words

    # REQ: Slave,Fn,AddrH,AddrL,NbWordsH,NbWordsL
    if slave_nb < 0 or slave_nb > 0xff:
        raise MbError(ErrorCode.BadData)

    if fn_nb not in (MbFunctionCode.MB_FN_READ_OUT_WORDS_3, MbFunctionCode.MB_FN_READ_IN_WORDS_4):
        raise MbError(ErrorCode.BadData)

    if addr < 0 or nb_words <= 0 or nb_words > MbLimits.MAX_READ_WORDS_NB or (addr + nb_words - 1) > 0xffff:
        raise MbError(ErrorCode.BadData)

    return _wrap_req(trans_id_none, bytes([slave_nb, fn_nb, hi(addr), lo(addr), 0, nb_words]))


# request: (trans_id, slave_nb, fn_nb, ...)
# resp:
#   None:   not enough data
#   MbException, MbError
#   f1, f2: (trans_id, slave_nb, fn_nb, (bit1: bool, bit2: bool, ...))
#   f3, f4: (trans_id, slave_nb, fn_nb, data_bytes)
def parse_res_tcp(ba: (bytes, bytearray), offset: int, nb: int, request=None) -> (None, tuple):
    assert isinstance(ba, (bytes, bytearray)), ba
    assert isinstance(offset, int), offset
    assert isinstance(nb, int), nb
    assert request is None or isinstance(request, tuple), request

    if nb < 6:
        return None

    rtu_size = word_ml(ba, offset + 4)
    tcp_size = rtu_size + 6

    if nb < tcp_size:
        return None

    if rtu_size < 2:
        raise MbError(ErrorCode.BadData)

    rtu_offset = 6
    trans_id = word_ml(ba, offset + 0)
    slave_nb = ba[rtu_offset + 0]
    fn_nb = ba[rtu_offset + 1] & 0x7f

    if request:
        r_trans_id, r_slave_nb, r_fn_nb, *r_data = request
        if r_trans_id != trans_id:  # Если запрос был rtu, r_trans_id=None; none != int
            raise MbError(ErrorCode.MismatchTrId)

        if r_slave_nb != slave_nb:
            raise MbError(ErrorCode.MismatchSlave)

        if r_fn_nb != fn_nb:
            raise MbError(ErrorCode.MismatchFn)

    if ba[rtu_offset + 1] & 0x80:
        if rtu_size != 3:
            raise MbError(ErrorCode.BadData)

        exc_nb = ba[rtu_offset + 2]

        raise MbException(exc_nb)

    if fn_nb in (MbFunctionCode.MB_FN_READ_OUT_BITS_1, MbFunctionCode.MB_FN_READ_IN_BITS_2):
        # REPLY: Slave,Fn,NbBytes,DATA
        if rtu_size < 3:
            raise MbError(ErrorCode.BadData)

        nb_bytes = ba[rtu_offset + 2]

        if rtu_size != (3 + nb_bytes) or nb_bytes == 0:
            raise MbError(ErrorCode.BadData)

        # nb_bytes = nb_bytes;
        # this->read_bits.Bytes	= RtuData + 3;

        if request:
            # r_addr = r_data[0]
            r_nb_bits = r_data[1]
            nb_bytes_requested = (r_nb_bits + 7) // 8
            if nb_bytes != nb_bytes_requested:
                raise MbError(ErrorCode.MismatchData)

        return tcp_size

    elif fn_nb in (MbFunctionCode.MB_FN_READ_OUT_WORDS_3, MbFunctionCode.MB_FN_READ_IN_WORDS_4):
        # REPLY: Slave,Fn,NbBytes,DATA
        if rtu_size < 3:
            raise MbError(ErrorCode.BadData)

        nb_bytes = ba[rtu_offset + 2]

        if rtu_size != (3 + nb_bytes) or nb_bytes == 0 or (nb_bytes % 2) != 0:
            raise MbError(ErrorCode.BadData)

        nb_words = nb_bytes // 2
        words_offs = rtu_offset + 3

        if request:
            # r_addr = r_data[0]
            r_nb_words = r_data[1]
            if nb_words != r_nb_words:
                raise MbError(ErrorCode.MismatchData)

        return trans_id, slave_nb, fn_nb, ba[words_offs:]

    elif fn_nb == MbFunctionCode.MB_FN_WRITE_BIT_5:
        # REPLY: Slave,Fn,AddrH,AddrL,Value(0/FF),0  - just a copy of request
        if rtu_size != 6:
            raise MbError(ErrorCode.BadData)

        addr = word_ml(ba, rtu_offset + 2)
        value = ba[rtu_offset + 4]

        if request:
            val = 0xff if request.write_bit.Value else 0x00
            if request.write_bit.addr != addr or val != value:
                raise MbError(ErrorCode.MismatchData)

        return tcp_size

    elif fn_nb == MbFunctionCode.MB_FN_WRITE_WORD_6:
        # REPLY: Slave,Fn,AddrH,AddrL,ValueH,ValueL - just a copy of request
        if rtu_size != 6:
            raise MbError(ErrorCode.BadData)

        addr = word_ml(ba, rtu_offset + 2)
        value = word_ml(ba, rtu_offset + 4)

        if request:
            if request.write_word.addr != addr or request.write_word.value != value:
                raise MbError(ErrorCode.MismatchData)

        return tcp_size

    elif fn_nb == MbFunctionCode.MB_FN_WRITE_BITS_15:
        # REPLY: Slave,Fn,AddrH,AddrL,NbBitsH,NbBitsL
        if rtu_size != 6:
            raise MbError(ErrorCode.BadData)

        addr = word_ml(ba, rtu_offset + 2)
        nb_bits = word_ml(ba, rtu_offset + 4)

        if request:
            if request.write_bits.addr != addr or request.write_bits.nb_bits != nb_bits:
                raise MbError(ErrorCode.MismatchData)

        return tcp_size

    elif fn_nb == MbFunctionCode.MB_FN_WRITE_WORDS_16:
        # REPLY: Slave,Fn,AddrH,AddrL,NbWordsH,NbWordsL
        if rtu_size != 6:
            raise MbError(ErrorCode.BadData)

        addr = word_ml(ba, rtu_offset + 2)
        nb_words = word_ml(ba, rtu_offset + 4)

        if request:
            if request.write_words.addr != addr or request.write_words.nb_words != nb_words:
                raise MbError(ErrorCode.MismatchData)

        return tcp_size
    else:
        assert False, fn_nb
        return None



#!/usr/bin/python3

from . import merk_nzif_ctx
from .nzif_defs import get_nzif_nack_codes


class NzifCtx(merk_nzif_ctx.MerkNzifCtx):
    def __init__(self, channel):
        super().__init__(channel, False)

    def get_nack_comment_n(self, nack_code):
        return get_nzif_nack_codes().get(nack_code, None)



#!/usr/bin/python3

'''
    Специфика каналов связана с их использованием в драйвере. Мне не нужна гора различных
    непредсказуемых Exception, потому что логика модуля подразумевает нормальную работу с ошибками В/В.
    В связи с этим все возможные ошибки должны быть сведены к одному исключению
        -   ConnectionError - при невозможности дальнейшего использования канала. При этом канал должен сам
                                закрываться. После повторного открытия канал снова можно использовать.
'''


class BaseChannel:
    # Doesnt' raise. Returns True if channel is connected
    def open(self, timeout_sec: float = 5.0, log=None) -> bool:
        raise NotImplementedError()

    # raise ConnectionError
    def reset_rcv_buf_x(self, log=None) -> None:
        self.read_x(0.0, log)

    # Doesnt' raise
    def safe_close(self, log=None) -> None:
        raise NotImplementedError()

    # Doesnt' raise
    def is_open(self) -> bool:
        raise NotImplementedError()

    # raise ConnectionError
    def write_all_x(self, data: (bytes, bytearray), log=None) -> None:
        raise NotImplementedError()

    # raise ConnectionError
    # if no data, returns empty bytes object
    def read_x(self, timeout_sec: float, log=None) -> bytes:
        raise NotImplementedError()



#!/usr/bin/python3

import argparse
import sys
import logging

from twisted.internet import reactor

from ..lib.log import configure_arg_parser_for_log, configure_logger
from ..lib import stdsvc
from ..lib.stddef import VOLCANO_DEFAULT_TCP_PORT
from ..lib.xml_reader import XmlReader, LoadException, load_xml_file_le, get_node_location

from ..twistedclient.twisted_factory import VolcanoTwistedFactory
from ..twistedclient.twisted_client_mt import VolcanoTwistedClientMT

from .channel_tcp import TcpChannel
from .channel_serial import SerialChannel
from . import merk_prot
from . import nzif_prot
from . import moxa_di_prot
from . import mb_prot
from . import tools
from . import build_cfg


def configure_my_args(parser):
    parser.add_argument('--core_host', help='Volcano core host', default='localhost')
    parser.add_argument('--core_port', help='Volcano core port', default=VOLCANO_DEFAULT_TCP_PORT, type=int)
    parser.add_argument('--core_connect_attempts', type=int)
    parser.add_argument('--core_reconnect_pause', type=int)

    parser.add_argument('-n', '--name', help='Instance name', default='poller')
    parser.add_argument('-f', '--file', help='Config file', default='poller.xml')
    parser.add_argument('-b', '--build', help='Build config', type=str, default='')


def load(env, log):
    cfg = load_xml_file_le(env.file)

    protocols_map = {}
    for node in cfg.getroot():
        p = XmlReader(node)

        if node.tag == 'Channel':
            # media
            media_type = node.attrib['channelType']
            if media_type == 'tcp':
                channel = TcpChannel(node)
            elif media_type == 'serial':
                channel = SerialChannel(node)
            else:
                raise LoadException('Unknown channel type: "{}". Available: {}'.format(media_type, ['tcp', 'serial']), node)

            # protocol
            proposed_aux_name = tools.unique_name('protocol_', lambda x: x not in protocols_map)

            protocol_name = p.get_str('protocol', allowed=('merk', 'nzif', 'modbus'))
            if protocol_name == 'merk':
                protocol = merk_prot.MerkProtocol(node, channel, proposed_aux_name, log)
            elif protocol_name == 'nzif':
                protocol = nzif_prot.NzifProtocol(node, channel, proposed_aux_name, log)
            else:
                protocol = mb_prot.MbProtocol(node, channel, proposed_aux_name, log)

            if not protocol.empty():
                protocols_map[protocol.aux_name()] = protocol
            else:
                log.warning('Protocol {} is empty and will be removed'.format(get_node_location(node)))
        
        elif node.tag == 'Moxa':

            # protocol
            proposed_aux_name = tools.unique_name('protocol_', lambda x: x not in protocols_map)

            protocol = moxa_di_prot.MoxaDIProtocol(node, proposed_aux_name, log)

            if not protocol.empty():
                protocols_map[protocol.aux_name()] = protocol
            else:
                log.warning('Protocol {} is empty and will be removed'.format(get_node_location(node)))
        else:
            raise LoadException('Unknown node name: "{}". Available: {}'.format(node.tag, ['Channel', 'Moxa']), node)

    log.info('Loaded {} channels'.format(len(protocols_map)))

    t_max = reactor.getThreadPool().max     # pylint: disable=no-member
    if len(protocols_map) > t_max:
        log.warning('Number of channels ({}) exceeds default thread pool size({}). Threading will be inefficient (some channels will wait for another to finish). Try splitting task into several processes.'.format(len(protocols_map), t_max))

    return protocols_map


class VolcanoClient(VolcanoTwistedClientMT):    # pylint: disable=abstract-method
    protocols_map = None

    def __init__(self, env):
        super().__init__(logging.getLogger(env.name))
        self.env = env
        self.services_ = None

    def connectionMade(self):
        assert self.protocols_map

        super().connectionMade()

        self.log.info('Connected to core')

        self.services_ = {
            'hs': stdsvc.HandshakeService(stdsvc.PipeMarker(self, 'hs')),
            'echo': stdsvc.SingleThreadEchoService(stdsvc.PipeMarker(self, 'echo')),
            'set': stdsvc.MultiThreadSetValueService(stdsvc.PipeMarker(self, 'set')),
            'find': stdsvc.SingleThreadFindTagService(stdsvc.PipeMarker(self, 'find')),
        }
        self.services_['hs'].salute(self.env.name)

        for proto in self.protocols_map.values():
            proto.sync(self.services_)
        self.services_['echo'].echo(self.on_sync_complete)

    def on_sync_complete(self):
        for proto in self.protocols_map.values():
            reactor.callInThread(proto.run_once_safe, reactor)  # pylint: disable=no-member

    def on_msg_rcvd_mt(self, msg: dict) -> None:
        assert 'ch' in msg, msg

        ch_name = msg['ch']
        assert ch_name in self.services_, (ch_name, self.services_)

        self.services_[ch_name].push_single_message(msg)  # error should raise


def main() -> int:
    try:
        # Args
        arg_parser = argparse.ArgumentParser()
        configure_my_args(arg_parser)
        configure_arg_parser_for_log(arg_parser)  # let logger add his arguments for later configure_logger() call
        env = arg_parser.parse_args()

        # Logging
        configure_logger(env)
        log = logging.getLogger(env.name)
        protocols_map = None

        try:
            if env.build:
                build_cfg.enable_build()
                load(env, log)
                build_cfg.flush_build()
                return 0
            else:
                protocols_map = load(env, log)

            factory = VolcanoTwistedFactory()
            factory.log = log
            factory.client = VolcanoClient(env)
            factory.client.protocols_map = protocols_map

            if env.core_connect_attempts is not None:
                factory.nb_attempts = env.core_connect_attempts

            if env.core_reconnect_pause is not None:
                factory.reconnect_pause = env.core_reconnect_pause

            log.info('Connect to core at %s:%s', env.core_host, env.core_port)
            reactor.connectTCP(env.core_host, env.core_port, factory)   # pylint: disable=no-member
            reactor.run()   # pylint: disable=no-member

            return 0

        except (Warning, LoadException) as e:
            log.error(e)
            return 1

        finally:
            # я не использую factory.stopFactory(), потому что она вызывается при каждой неудачной попытке подключения к ядру
            if protocols_map:
                for proto in protocols_map.values():
                    proto.safe_close()
    finally:
        logging.shutdown()


if __name__ == '__main__':
    sys.exit(main())
